# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from decimal import Decimal
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
import abc
import collections
import logging

from basana.backtesting import errors
from basana.backtesting.exchange import Exchange
from basana.core import bar, event, helpers
from basana.core.enums import OrderOperation
from basana.core.pair import Pair

import plotly.graph_objects as go  # type: ignore
import plotly.subplots  # type: ignore


ChartDataPointFn = Callable[[datetime], Optional[Decimal]]
logger = logging.getLogger(__name__)


class DataPointFromSequence:
    """Callable that returns the last value of a sequence if its not empty.

    :param seq: The sequence that will be used to get the value.
    """
    def __init__(self, seq: Sequence[Any]):
        self._seq = seq

    def __call__(self, dt: datetime) -> Optional[Decimal]:
        ret = None
        if self._seq:
            ret = self._seq[-1]
        return Decimal(ret) if ret is not None else ret


class TimeSeries:
    def __init__(self):
        self._values = {}

    def add_value(self, dt: datetime, value: Decimal):
        self._values[dt] = value

    def get_x_y(self):
        return zip(*sorted(self._values.items())) if self._values else ([], [])


class LineChart(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_title(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def add_traces(self, figure: go.Figure, row: int):
        raise NotImplementedError()


class PairLineChart(LineChart):
    def __init__(self, pair: Pair, include_buys: bool, include_sells: bool, exchange: Exchange):
        self._pair = pair
        self._include_buys = include_buys
        self._include_sells = include_sells
        self._exchange = exchange
        self._ts = TimeSeries()
        self._indicators: Dict[str, Tuple[ChartDataPointFn, TimeSeries]] = {}

        exchange.subscribe_to_bar_events(pair, self._on_bar_event)

    def get_title(self) -> str:
        return str(self._pair)

    def add_traces(self, figure: go.Figure, row: int):
        # Add a trace with the pair values.
        x, y = self._ts.get_x_y()
        figure.add_trace(go.Scatter(x=x, y=y, name=str(self._pair)), row=row, col=1)

        # Add a trace with buy prices.
        if self._include_buys:
            x, y = self._get_order_fills(OrderOperation.BUY).get_x_y()
            figure.add_trace(
                go.Scatter(x=x, y=y, name="Buy", mode="markers", marker=dict(symbol="arrow-up", color="green")),
                row=row, col=1
            )

        # Add a trace with sell prices.
        if self._include_sells:
            x, y = self._get_order_fills(OrderOperation.SELL).get_x_y()
            figure.add_trace(
                go.Scatter(x=x, y=y, name="Sell", mode="markers", marker=dict(symbol="arrow-down", color="red")),
                row=row, col=1
            )

        # Add one trace per indicator.
        for name, (_, ts) in self._indicators.items():
            x, y = ts.get_x_y()
            figure.add_trace(go.Scatter(x=x, y=y, name=name), row=row, col=1)

    def add_indicator(self, name: str, get_data_point: ChartDataPointFn):
        assert name not in self._indicators
        self._indicators[name] = (get_data_point, TimeSeries())

    def _get_order_fills(self, op: OrderOperation) -> TimeSeries:
        ret = TimeSeries()
        orders = filter(
            lambda order: order.pair == self._pair and order.operation == op,
            self._exchange._get_all_orders()
        )
        pair_info = self._exchange._get_pair_info(self._pair)
        for order in orders:
            for fill in order.fills:
                base_amount = fill.balance_updates[order.pair.base_symbol]
                quote_amount = fill.balance_updates[order.pair.quote_symbol]
                price = -helpers.truncate_decimal(quote_amount / base_amount, pair_info.quote_precision)
                ret.add_value(fill.when, price)
        return ret

    async def _on_bar_event(self, bar_event: bar.BarEvent):
        dt = bar_event.when
        value = bar_event.bar.close
        # Add the value to the main time series.
        self._ts.add_value(dt, value)
        # Add the custom indicator values.
        for get_data_point, ts in self._indicators.values():
            indicator_value = get_data_point(dt)
            if indicator_value is not None:
                ts.add_value(dt, indicator_value)


class AccountBalanceLineChart(LineChart):
    def __init__(self, symbol: str, exchange: Exchange):
        self._symbol = symbol
        self._exchange = exchange
        self._ts = TimeSeries()

        # Initially I thought of having the exchange emit an event when any balance got updated, but I then realized
        # that it would be too much overhead if charts are not used.
        exchange._get_dispatcher().subscribe_all(self._on_any_event)

    def get_title(self) -> str:
        return f"{self._symbol} balance"

    def add_traces(self, figure: go.Figure, row: int):
        # Add a trace with the pair values.
        x, y = self._ts.get_x_y()
        figure.add_trace(go.Scatter(x=x, y=y, name=self._symbol), row=row, col=1)

    async def _on_any_event(self, event: event.Event):
        balance = await self._exchange.get_balance(self._symbol)
        self._ts.add_value(event.when, balance.total - balance.borrowed)


class PortfolioValueLineChart(LineChart):
    def __init__(self, symbol: str, exchange: Exchange, precision: int = 2):
        self._symbol = symbol
        self._exchange = exchange
        self._ts = TimeSeries()
        self._precision = precision

        # Initially I thought of having the exchange emit an event when any balance got updated, but I then realized
        # that it would be too much overhead if charts are not used.
        exchange._get_dispatcher().subscribe_all(self._on_any_event)

    def get_title(self) -> str:
        return f"Portfolio value in {self._symbol}"

    def add_traces(self, figure: go.Figure, row: int):
        # Add a trace with the portfolio values.
        x, y = self._ts.get_x_y()
        figure.add_trace(go.Scatter(x=x, y=y, name=f"Portfolio ({self._symbol})"), row=row, col=1)

    async def _on_any_event(self, event: event.Event):
        portfolio_value = Decimal(0)
        balances = await self._exchange.get_balances()
        for symbol, balance in balances.items():
            if balance.total == 0:
                continue

            try:
                rate: Decimal = Decimal(1)
                if symbol != self._symbol:
                    rate, _ = await self._exchange.get_bid_ask(Pair(symbol, self._symbol))
                portfolio_value += rate * balance.total
            except errors.Error as e:
                logger.debug(str(e))

        self._ts.add_value(event.when, helpers.round_decimal(portfolio_value, self._precision))


class CustomLineChart(LineChart):
    def __init__(self, name: str, exchange: Exchange):
        self._name = name
        self._exchange = exchange
        self._data_point_fns: Dict[str, Tuple[ChartDataPointFn, TimeSeries]] = {}

        exchange._get_dispatcher().subscribe_all(self._on_any_event)

    def get_title(self) -> str:
        return self._name

    def add_traces(self, figure: go.Figure, row: int):
        for name, (_, ts) in self._data_point_fns.items():
            x, y = ts.get_x_y()
            figure.add_trace(go.Scatter(x=x, y=y, name=name), row=row, col=1)

    def add_data_point_fn(self, name: str, get_data_point: ChartDataPointFn):
        assert name not in self._data_point_fns
        self._data_point_fns[name] = (get_data_point, TimeSeries())

    async def _on_any_event(self, event: event.Event):
        dt = event.when
        for get_data_point, ts in self._data_point_fns.values():
            value = get_data_point(dt)
            if value is not None:
                ts.add_value(dt, value)


class LineCharts:
    """A set of line charts that show the evolution of pair prices and account balances over time.

    :param exchange: The backtesting exchange.
    """
    def __init__(self, exchange: Exchange):
        self._exchange = exchange
        self._balance_charts: Dict[str, AccountBalanceLineChart] = collections.OrderedDict()
        self._pair_charts: Dict[Pair, PairLineChart] = collections.OrderedDict()
        self._portfolio_charts: Dict[str, PortfolioValueLineChart] = collections.OrderedDict()
        self._custom_charts: Dict[str, CustomLineChart] = collections.OrderedDict()

    def add_balance(self, symbol: str):
        """Adds a chart with an account's balance.

        :param symbol: The currency symbol.
        """
        self._balance_charts[symbol] = AccountBalanceLineChart(symbol, self._exchange)

    def add_portfolio_value(self, symbol: str, precision: int = 2):
        """Adds a chart with the portfolio value in a given currency.

        :param symbol: The currency symbol.
        :param precision: The number of digits after the decimal point.

        .. note::

            * If the portfolio value can't be calculated at any given point, for example because there is no price for
              a given instrument, an error will be logged.
        """
        self._portfolio_charts[symbol] = PortfolioValueLineChart(symbol, self._exchange, precision=precision)

    def add_pair(self, pair: Pair, include_buys: bool = True, include_sells: bool = True):
        """Adds a chart with the pair values.

        :param pair: The pair.
        :param include_buys: True to include buy prices.
        :param include_sells: True to include sell prices.
        """
        self._pair_charts[pair] = PairLineChart(pair, include_buys, include_sells, self._exchange)

    def add_pair_indicator(self, name: str, pair: Pair, get_data_point: ChartDataPointFn):
        """Adds a technical indicator to a pair's chart.

        :param name: The name of the indicator.
        :param pair: The pair chart to add the indicator to.
        :param get_data_point: A callable that will be used to get the data point on each bar.
        """
        assert pair in self._pair_charts, f"{pair} was not added"
        self._pair_charts[pair].add_indicator(name, get_data_point)

    def add_custom(self, name: str, line: str, get_data_point: ChartDataPointFn):
        """Adds a custom chart.

        :param name: The name of the chart.
        :param line: The name of the line.
        :param get_data_point: A callable that will be used to get the line data points.
        """
        if (chart := self._custom_charts.get(name)) is None:
            chart = CustomLineChart(name, self._exchange)
            self._custom_charts[name] = chart
        chart.add_data_point_fn(line, get_data_point)

    def show(self, show_legend: bool = True):  # pragma: no cover
        """Shows the chart using either the default renderer(s).

        Check https://plotly.com/python-api-reference/generated/plotly.graph_objects.Figure.html#plotly.graph_objects.Figure.show
        for more information.

        :param show_legend: True if legends should be visible, False otherwise.
        """  # noqa: E501

        if fig := self._build_figure(show_legend=show_legend):
            fig.show()

    def save(
            self, path: str, width: Optional[int] = None, height: Optional[int] = None,
            scale: Optional[Union[int, float]] = None, show_legend: bool = True
    ):
        """Saves the chart to a file.

        :param path: The path to the file to save the image.
        :param width: The width of the exported image in layout pixels.
        :param height: The height of the exported image in layout pixels.
        :param scale: The scale factor to use when exporting the figure.
        :param show_legend: True if legends should be visible, False otherwise.

        .. note::

            * The supported file formats are png, jpg/jpeg, webp, svg and pdf.
        """

        if fig := self._build_figure(show_legend=show_legend):
            fig.write_image(path, width=width, height=height, scale=scale)

    def _build_figure(self, show_legend: bool = True) -> Optional[go.Figure]:
        charts: List[LineChart] = []
        charts.extend(self._pair_charts.values())
        charts.extend(self._balance_charts.values())
        charts.extend(self._portfolio_charts.values())
        charts.extend(self._custom_charts.values())

        figure = None
        if charts:
            subplot_titles = [chart.get_title() for chart in charts]
            figure = plotly.subplots.make_subplots(
                rows=len(charts), cols=1, shared_xaxes=True, subplot_titles=subplot_titles
            )

            row = 1
            for chart in charts:
                chart.add_traces(figure, row)
                row += 1

            figure.layout.update(showlegend=show_legend)

        return figure



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Dict, List, Optional


from . import base


# https://binance-docs.github.io/apidocs/spot/en/#spot-account-trade
class SpotAccount:
    def __init__(self, client: base.BaseClient):
        self._client = client

    async def get_account_information(self) -> dict:
        return await self._client.make_request("GET", "/api/v3/account", send_sig=True)

    async def create_order(
            self, symbol: str, side: str, type: str, time_in_force: Optional[str] = None,
            quantity: Optional[Decimal] = None, quote_order_qty: Optional[Decimal] = None,
            price: Optional[Decimal] = None, stop_price: Optional[Decimal] = None,
            new_client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> dict:
        params: Dict[str, Any] = {
            "symbol": symbol,
            "side": side,
            "type": type,
        }
        base.set_optional_params(params, (
            ("timeInForce", time_in_force),
            ("quantity", quantity),
            ("quoteOrderQty", quote_order_qty),
            ("price", price),
            ("stopPrice", stop_price),
            ("newClientOrderId", new_client_order_id),
        ))
        params.update(kwargs)
        return await self._client.make_request("POST", "/api/v3/order", data=params, send_sig=True)

    async def query_order(
            self, symbol: str, order_id: Optional[int] = None, orig_client_order_id: Optional[str] = None
    ) -> dict:
        assert (order_id is not None) ^ (orig_client_order_id is not None), \
            "Either order_id or orig_client_order_id should be set"

        params: Dict[str, Any] = {"symbol": symbol}
        base.set_optional_params(params, (
            ("orderId", order_id),
            ("origClientOrderId", orig_client_order_id),
        ))
        return await self._client.make_request("GET", "/api/v3/order", qs_params=params, send_sig=True)

    async def get_open_orders(
            self, symbol: Optional[str] = None
    ) -> dict:
        params: Dict[str, Any] = {}
        if symbol is not None:
            params["symbol"] = symbol
        return await self._client.make_request("GET", "/api/v3/openOrders", qs_params=params, send_sig=True)

    async def cancel_order(
            self, symbol: str, order_id: Optional[int] = None, orig_client_order_id: Optional[str] = None
    ) -> dict:
        assert (order_id is not None) ^ (orig_client_order_id is not None), \
            "Either order_id or orig_client_order_id should be set"

        params: Dict[str, Any] = {"symbol": symbol}
        base.set_optional_params(params, (
            ("orderId", order_id),
            ("origClientOrderId", orig_client_order_id),
        ))
        return await self._client.make_request("DELETE", "/api/v3/order", qs_params=params, send_sig=True)

    async def get_trades(self, symbol: str, order_id: Optional[int] = None) -> List[dict]:
        params: Dict[str, Any] = {"symbol": symbol}
        if order_id is not None:
            params["orderId"] = order_id
        return await self._client.make_request("GET", "/api/v3/myTrades", qs_params=params, send_sig=True)

    async def create_oco(
            self, symbol: str, side: str, quantity: Decimal, price: Decimal, stop_price: Decimal,
            stop_limit_price: Optional[Decimal] = None, stop_limit_time_in_force: Optional[str] = None,
            list_client_order_id: Optional[str] = None, limit_client_order_id: Optional[str] = None,
            stop_client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> dict:
        params: Dict[str, Any] = {
            "symbol": symbol,
            "side": side,
            "quantity": str(quantity),
            "price": str(price),
            "stopPrice": str(stop_price),
        }
        base.set_optional_params(params, (
            ("listClientOrderId", list_client_order_id),
            ("stopLimitPrice", stop_limit_price),
            ("stopLimitTimeInForce", stop_limit_time_in_force),
            ("limitClientOrderId", limit_client_order_id),
            ("stopClientOrderId", stop_client_order_id),
        ))
        params.update(kwargs)
        return await self._client.make_request("POST", "/api/v3/order/oco", data=params, send_sig=True)

    async def cancel_oco_order(
            self, symbol: str, order_list_id: Optional[int] = None, client_order_list_id: Optional[str] = None
    ) -> dict:
        assert (order_list_id is not None) ^ (client_order_list_id is not None), \
            "Either order_list_id or client_order_list_id should be set"

        params: Dict[str, Any] = {
            "symbol": symbol
        }
        base.set_optional_params(params, (
            ("orderListId", order_list_id),
            ("origClientOrderId", client_order_list_id),
        ))
        return await self._client.make_request("DELETE", "/api/v3/orderList", data=params, send_sig=True)

    async def query_oco_order(
            self, order_list_id: Optional[int] = None, client_order_list_id: Optional[str] = None
    ) -> dict:
        assert (order_list_id is not None) ^ (client_order_list_id is not None), \
            "Either order_list_id or client_order_list_id should be set"

        params: Dict[str, Any] = {}
        base.set_optional_params(params, (
            ("orderListId", order_list_id),
            ("origClientOrderId", client_order_list_id),
        ))
        return await self._client.make_request("GET", "/api/v3/orderList", qs_params=params, send_sig=True)

    async def create_listen_key(self) -> dict:
        return await self._client.make_request("POST", "/api/v3/userDataStream", send_key=True)

    async def keep_alive_listen_key(self, listen_key: str) -> dict:
        params: Dict[str, Any] = {
            "listenKey": listen_key,
        }
        return await self._client.make_request("PUT", "/api/v3/userDataStream", send_key=True, data=params)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import cast, Callable, Optional

import aiohttp

from . import client, order_book, trades, user_data, websockets as binance_ws, klines
from basana.core import bar, dispatcher, websockets as core_ws
from basana.core.pair import Pair


class WebsocketManager:
    def __init__(
            self, dispatcher: dispatcher.EventDispatcher, api_client: client.APIClient,
            session: Optional[aiohttp.ClientSession] = None, config_overrides: dict = {}
    ):
        self._dispatcher = dispatcher
        self._cli = api_client
        self._session = session
        self._config_overrides = config_overrides
        self._websocket: Optional[binance_ws.WebSocketClient] = None

    def subscribe_to_bar_events(self, pair: Pair, interval: str, event_handler: bar.BarEventHandler):
        self._subscribe_to_ws_channel_events(
            binance_ws.PublicChannel(klines.get_channel(pair, interval)),
            lambda ws_cli: klines.WebSocketEventSource(pair, ws_cli),
            cast(dispatcher.EventHandler, event_handler)
        )

    def subscribe_to_order_book_events(
            self, pair: Pair, event_handler: order_book.OrderBookEventHandler, depth: int = 10
    ):
        self._subscribe_to_ws_channel_events(
            binance_ws.PublicChannel(order_book.get_channel(pair, depth)),
            lambda ws_cli: order_book.WebSocketEventSource(pair, ws_cli),
            cast(dispatcher.EventHandler, event_handler)
        )

    def subscribe_to_trade_events(self, pair: Pair, event_handler: trades.TradeEventHandler):
        self._subscribe_to_ws_channel_events(
            binance_ws.PublicChannel(trades.get_channel(pair)),
            lambda ws_cli: trades.WebSocketEventSource(pair, ws_cli),
            cast(dispatcher.EventHandler, event_handler)
        )

    def subscribe_to_user_data_events(
        self, channel: binance_ws.Channel,
        event_src_factory: Callable[[core_ws.WebSocketClient], core_ws.ChannelEventSource],
        event_handler: user_data.UserDataEventHandler,
    ):
        self._subscribe_to_ws_channel_events(
            channel, event_src_factory, cast(dispatcher.EventHandler, event_handler)
        )

    def subscribe_to_order_events(
        self, channel: binance_ws.Channel,
        event_src_factory: Callable[[core_ws.WebSocketClient], core_ws.ChannelEventSource],
        event_handler: user_data.OrderEventHandler,
    ):
        async def forward_if_order_event(event: user_data.Event):
            if isinstance(event, user_data.OrderEvent):
                await event_handler(event)

        self._subscribe_to_ws_channel_events(
            channel, event_src_factory, cast(dispatcher.EventHandler, forward_if_order_event)
        )

    def _subscribe_to_ws_channel_events(
            self, channel: binance_ws.Channel,
            event_src_factory: Callable[[core_ws.WebSocketClient], core_ws.ChannelEventSource],
            event_handler: dispatcher.EventHandler
    ):
        # Get/create the event source for the channel.
        ws_cli = self._get_ws_client()
        event_source = ws_cli.get_channel_event_source_ex(channel)
        if not event_source:
            event_source = event_src_factory(ws_cli)
            ws_cli.set_channel_event_source_ex(channel, event_source)

        # Subscribe the event handler to the event source.
        self._dispatcher.subscribe(event_source, event_handler)

    def _get_ws_client(self) -> binance_ws.WebSocketClient:
        if self._websocket is None:
            self._websocket = binance_ws.WebSocketClient(
                self._dispatcher, self._cli, session=self._session, config_overrides=self._config_overrides
            )
        return self._websocket



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Coroutine, List, Optional, Set, Union
import asyncio
import contextlib
import decimal
import logging
import warnings

import aiohttp

from basana.core import logs


class TaskGroup:
    def __init__(self):
        self._tasks = []
        self._exiting = False

    async def __aenter__(self) -> "TaskGroup":
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        self._exiting = True

        try:
            if not exc_type:
                await asyncio.gather(*self._tasks)
        finally:
            pending = self._cancel()
            if pending:
                # Don't raise exceptions since we're waiting for tasks to finish.
                await asyncio.gather(*pending, return_exceptions=True)

    def _cancel(self) -> List[asyncio.Task]:
        pending = [task for task in self._tasks if not task.done()]
        for task in pending:
            if not task.done():
                task.cancel()
        return pending

    def create_task(self, coro) -> asyncio.Task:
        assert not self._exiting
        ret = asyncio.create_task(coro)
        self._tasks.append(ret)
        return ret

    def cancel(self):
        self._cancel()


class TaskPool:
    """
    A class for managing a pool of asyncio tasks.

    :param size: The maximum size of the task pool.
    """
    def __init__(self, size: int):
        assert size > 0, "Invalid size"
        self._max_size = size
        self._tasks: Set[asyncio.Task] = set()
        self._done: List[asyncio.Task] = []

    @property
    def idle(self) -> int:
        """
        True if there are no active tasks in the pool, False otherwise.
        """
        return len(self._tasks) == 0

    async def push(self, coroutine: Coroutine[Any, Any, Any]):
        """
        Adds a coroutine to the task pool. If the pool is full it will block until there is room for the new task.

        :param coroutine: The coroutine to be added to the task pool.
        """
        # Wait for some task to complete if there is no more room.
        while len(self._tasks) >= self._max_size:
            await self._wait_impl(timeout=None, return_when=asyncio.FIRST_COMPLETED)
        self._tasks.add(asyncio.create_task(coroutine))

    def pop_done(self) -> List[asyncio.Task]:
        """
        Returns the tasks that are done running since the last call to pop_done.
        """
        ret = self._done
        self._done = []
        return ret

    def cancel(self):
        """
        Requests all tasks in the pool to be canceled.
        """
        pending = [task for task in self._tasks if not task.done()]
        for task in pending:
            task.cancel()

    async def wait(self, timeout: Optional[Union[int, float]] = None) -> bool:
        """
        Waits for all tasks in the pool to complete. Returns True if all tasks are done, False otherwise.

        :param timeout: The maximum number of seconds to wait for tasks to complete. If None, wait indefinitely.
        """
        return await self._wait_impl(timeout=timeout, return_when=asyncio.ALL_COMPLETED)

    async def _wait_impl(self, timeout: Optional[Union[int, float]], return_when: str) -> bool:
        done: List[asyncio.Task] = []
        if self._tasks:
            done, _ = await asyncio.wait(self._tasks, timeout=timeout, return_when=return_when)
        for task in done:
            self._tasks.remove(task)
            self._done.append(task)
        return len(done) > 0


@contextlib.contextmanager
def no_raise(logger: logging.Logger, msg: str, **kwargs):
    try:
        yield
    except Exception as e:
        log_args = {"exception": e}
        log_args.update(kwargs)
        logger.exception(logs.StructuredMessage(msg, **log_args))


@contextlib.asynccontextmanager
async def use_or_create_session(session: Optional[aiohttp.ClientSession] = None):
    if session:
        yield session
    else:
        async with aiohttp.ClientSession() as new_session:
            yield new_session


def round_decimal(value: Decimal, precision: int, rounding=None) -> Decimal:
    """Rounds a decimal value.

    :param value: The value to round.
    :param precision: The number of digits after the decimal point.
    :param rounding: An optional rounding option from the :mod:`decimal` module.
    :returns: The rounded value.
    """
    return value.quantize(Decimal(f"1e-{precision}"), rounding=rounding)


def truncate_decimal(value: Decimal, precision: int) -> Decimal:
    """Truncates a decimal value.

    :param value: The value to truncate.
    :param precision: The number of digits after the decimal point.
    :returns: The truncated value.
    """
    return round_decimal(value, precision, rounding=decimal.ROUND_DOWN)


def deprecation_warning(message: str):
    warnings.warn(message, DeprecationWarning, stacklevel=2)


def classpath(obj: object):
    cls = obj.__class__
    module = cls.__module__
    parts = [str(module), cls.__qualname__] if module else [cls.__qualname__]
    return ".".join(parts)


class LazyProxy:
    def __init__(self, factory):
        self._factory = factory
        self._obj = None

    @property
    def initialized(self):
        return self._obj is not None

    @property
    def obj(self):
        if self._obj is None:
            self._obj = self._factory()
        return self._obj

    def __getattr__(self, name):
        return getattr(self.obj, name)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Dict
import itertools


from basana.backtesting import config
from basana.core import helpers


ZERO = Decimal(0)
ValueMapDict = Dict[str, Decimal]


class ValueMap(ValueMapDict):
    def prune(self):
        keys = [key for key, value in self.items() if not value]
        for key in keys:
            del self[key]

    def truncate(self, config: config.Config):
        for symbol, amount in self.items():
            symbol_info = config.get_symbol_info(symbol)
            self[symbol] = helpers.truncate_decimal(amount, symbol_info.precision)

    def __add__(self, other: ValueMapDict) -> "ValueMap":
        keys = set(itertools.chain(self.keys(), other.keys()))
        return ValueMap({key: self.get(key, ZERO) + other.get(key, ZERO) for key in keys})

    def __iadd__(self, other: ValueMapDict) -> "ValueMap":
        keys = set(itertools.chain(self.keys(), other.keys()))
        for key in keys:
            self[key] = self.get(key, ZERO) + other.get(key, ZERO)
        return self

    def __radd__(self, other: ValueMapDict) -> "ValueMap":
        return self.__add__(other)

    def __sub__(self, other: ValueMapDict) -> "ValueMap":
        keys = set(itertools.chain(self.keys(), other.keys()))
        return ValueMap({key: self.get(key, ZERO) - other.get(key, ZERO) for key in keys})

    def __isub__(self, other: ValueMapDict) -> "ValueMap":
        keys = set(itertools.chain(self.keys(), other.keys()))
        for key in keys:
            self[key] = self.get(key, ZERO) - other.get(key, ZERO)
        return self

    def __rsub__(self, other: ValueMapDict) -> "ValueMap":
        keys = set(itertools.chain(self.keys(), other.keys()))
        return ValueMap({key: other.get(key, ZERO) - self.get(key, ZERO) for key in keys})

    def __mul__(self, other: ValueMapDict) -> "ValueMap":
        keys = set(itertools.chain(self.keys(), other.keys()))
        ret = ValueMap()
        for key in keys:
            ret[key] = self.get(key, ZERO) * other.get(key, ZERO)
        return ret

    def __imul__(self, other: ValueMapDict) -> "ValueMap":
        keys = set(itertools.chain(self.keys(), other.keys()))
        for key in keys:
            self[key] = self.get(key, ZERO) * other.get(key, ZERO)
        return self

    def __rmul__(self, other: ValueMapDict) -> "ValueMap":
        return self.__mul__(other)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

DEFAULTS = {
    "api": {
        "http": {
            "base_url": "https://api.binance.com/",
            "timeout": 30,
        },
        "websockets": {
            "base_url": "wss://stream.binance.com/",
            "heartbeat": 30,
            "spot": {
                "user_data_stream": {
                    "heartbeat": 15 * 60,
                },
            },
            "cross_margin": {
                "user_data_stream": {
                    "heartbeat": 15 * 60,
                },
            },
            "isolated_margin": {
                "user_data_stream": {
                    "heartbeat": 15 * 60,
                },
            },
        }
    }
}



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Dict, List, Optional
import copy

from basana.backtesting import errors, helpers as bt_helpers
from basana.backtesting.value_map import ValueMap
from basana.backtesting.lending import base as lending_base


class LoanManager:
    def __init__(
            self, lending_strategy: lending_base.LendingStrategy, exchange_ctx: lending_base.ExchangeContext
    ):
        self._loans = bt_helpers.ExchangeObjectContainer[lending_base.Loan]()
        self._ctx = exchange_ctx
        self._lending_strategy = lending_strategy
        self._collateral_by_loan: Dict[str, ValueMap] = {}
        self._lending_strategy.set_exchange_context(self, exchange_ctx)

    def create_loan(self, symbol: str, amount: Decimal) -> lending_base.LoanInfo:
        if amount <= 0:
            raise errors.Error("Invalid amount")

        # Create the loan and update balances.
        loan = self._lending_strategy.create_loan(symbol, amount, self._ctx.dispatcher.now())
        required_collateral = loan.calculate_collateral(self._ctx.prices)
        self._ctx.account_balances.update(
            balance_updates={loan.borrowed_symbol: loan.borrowed_amount},
            borrowed_updates={loan.borrowed_symbol: loan.borrowed_amount},
            hold_updates=required_collateral
        )

        # Save the loan now that balance updates succeeded.
        self._loans.add(loan)
        self._collateral_by_loan[loan.id] = ValueMap(required_collateral)

        return self._build_loan_info(loan)

    def get_loans(
            self, borrowed_symbol: Optional[str] = None, is_open: Optional[bool] = None
    ) -> List[lending_base.LoanInfo]:
        loans = self._loans.get_all()
        if borrowed_symbol:
            loans = filter(lambda loan: loan.borrowed_symbol == borrowed_symbol, loans)
        if is_open is not None:
            loans = filter(lambda loan: loan.is_open == is_open, loans)
        return [self._build_loan_info(loan) for loan in loans]

    def get_loan(self, loan_id: str) -> Optional[lending_base.LoanInfo]:
        loan = self._loans.get(loan_id)
        return None if loan is None else self._build_loan_info(loan)

    def repay_loan(self, loan_id: str):
        loan = self._get_open_loan(loan_id)

        interest = ValueMap()
        interest += loan.calculate_interest(self._ctx.dispatcher.now(), self._ctx.prices)
        interest.truncate(self._ctx.config)
        interest.prune()
        collateral = self._collateral_by_loan[loan_id]

        # Update balances.
        balance_updates = ValueMap({loan.borrowed_symbol: -loan.borrowed_amount})
        balance_updates -= interest
        self._ctx.account_balances.update(
            balance_updates=balance_updates,
            borrowed_updates={loan.borrowed_symbol: -loan.borrowed_amount},
            hold_updates={symbol: -amount for symbol, amount in collateral.items()}
        )

        # Close the loan now that balance updates succeeded.
        loan.add_paid_interest(interest)
        loan.close()
        self._collateral_by_loan.pop(loan_id)

    def cancel_loan(self, loan_id: str):
        loan = self._get_open_loan(loan_id)

        # Only loans that have just been created can be canceled
        assert loan.created_at == self._ctx.dispatcher.now()

        # Update balances.
        collateral = self._collateral_by_loan[loan_id]
        balance_updates = ValueMap({loan.borrowed_symbol: -loan.borrowed_amount})
        self._ctx.account_balances.update(
            balance_updates=balance_updates,
            borrowed_updates=balance_updates,
            hold_updates={symbol: -amount for symbol, amount in collateral.items()}
        )

        # Close the loan now that balance updates succeeded.
        loan.close()
        self._collateral_by_loan.pop(loan_id)

    def _get_open_loan(self, loan_id: str) -> lending_base.Loan:
        loan = self._loans.get(loan_id)
        if not loan:
            raise errors.NotFound("Loan not found")
        if not loan.is_open:
            raise errors.Error("Loan is not open")
        return loan

    def _build_loan_info(self, loan: lending_base.Loan) -> lending_base.LoanInfo:
        outstanding_interest = ValueMap()
        if loan.is_open:
            outstanding_interest += loan.calculate_interest(
                self._ctx.dispatcher.now(), self._ctx.prices
            )
            outstanding_interest.truncate(self._ctx.config)
            outstanding_interest.prune()

        return lending_base.LoanInfo(
            id=loan.id, is_open=loan.is_open, borrowed_symbol=loan.borrowed_symbol,
            borrowed_amount=loan.borrowed_amount, outstanding_interest=outstanding_interest,
            paid_interest=copy.copy(loan.paid_interest)
        )



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses


@dataclasses.dataclass(frozen=True)
class Pair:
    """A trading pair.

    :param base_symbol: The base symbol. It could be a stock, a crypto currency, a currency, etc.
    :param quote_symbol: The quote symbol. It could be a stock, a crypto currency, a currency, etc.
    """

    #: The base symbol.
    base_symbol: str

    #: The quote symbol.
    quote_symbol: str

    def __str__(self):
        return "{}/{}".format(self.base_symbol, self.quote_symbol)


@dataclasses.dataclass(frozen=True)
class PairInfo:
    """Information about a trading pair.

    :param base_precision: The precision for the base symbol.
    :param quote_precision: The precision for the quote symbol.
    """

    #: The precision for the base symbol.
    base_precision: int

    #: The precision for the quote symbol.
    quote_precision: int



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Awaitable, cast, Callable, Dict, Generator, Iterable, List, Optional
import dataclasses
import datetime
import decimal
import logging

from basana.backtesting import account_balances, config, errors, fees, helpers, lending, loan_mgr, liquidity, prices
from basana.backtesting.orders import Order, OrderInfo
from basana.backtesting.value_map import ValueMap, ValueMapDict
from basana.core import bar, dispatcher, helpers as core_helpers, logs
from basana.core.enums import OrderOperation
from basana.core.pair import Pair
import basana as bs


logger = logging.getLogger(__name__)

LiquidityStrategyFactory = Callable[[], liquidity.LiquidityStrategy]
OrderEventHandler = Callable[["OrderEvent"], Awaitable[Any]]


@dataclasses.dataclass
class ExchangeContext:
    dispatcher: dispatcher.BacktestingDispatcher
    account_balances: account_balances.AccountBalances
    prices: prices.Prices
    fee_strategy: fees.FeeStrategy
    liquidity_strategy_factory: LiquidityStrategyFactory
    loan_mgr: loan_mgr.LoanManager
    config: config.Config


class OrderEvent(bs.Event):
    """
    An event for order updates.
    """

    def __init__(self, when: datetime.datetime, order: OrderInfo):
        super().__init__(when)
        #: The order.
        self.order: OrderInfo = order


class OrderManager:
    def __init__(self, exchange_ctx: ExchangeContext):
        self._ctx = exchange_ctx
        self._liquidity_strategies: Dict[Pair, liquidity.LiquidityStrategy] = {}
        self._orders = helpers.ExchangeObjectContainer[Order]()
        self._holds_by_order: Dict[str, ValueMap] = {}
        self._order_updates = core_helpers.LazyProxy(bs.FifoQueueEventSource)

    def on_bar_event(self, bar_event: bar.BarEvent):
        if (liquidity_strategy := self._liquidity_strategies.get(bar_event.bar.pair)) is None:
            liquidity_strategy = self._ctx.liquidity_strategy_factory()
        liquidity_strategy.on_bar(bar_event.bar)
        for order in filter(lambda o: o.pair == bar_event.bar.pair, self._orders.get_open()):
            self._process_order(order, bar_event, liquidity_strategy)

    def add_order(self, order: Order):
        try:
            # When an order gets accepted we need to hold any required balance that will be debited as the order gets
            # filled.
            if required_balances := self._estimate_required_balances(order):
                if order.auto_borrow:
                    self._borrow(required_balances, order)

                self._ctx.account_balances.update(hold_updates=required_balances)
                self._holds_by_order[order.id] = required_balances

            # The order got accepted.
            self._orders.add(order)
            self._push_order_update(order)

        except errors.NotEnoughBalance as e:
            logger.debug(logs.StructuredMessage(
                "Not enough balance to accept order", order=order.get_debug_info(), error=str(e)
            ))
            raise

    def get_order(self, order_id: str) -> Optional[Order]:
        return self._orders.get(order_id)

    def get_all_orders(self) -> Iterable[Order]:
        return self._orders.get_all()

    def get_open_orders(self) -> Generator[Order, None, None]:
        return self._orders.get_open()

    def cancel_order(self, order_id: str):
        order = self._orders.get(order_id)
        if order is None:
            raise errors.Error("Order not found")
        if not order.is_open:
            raise errors.Error("Order {} is in {} state and can't be canceled".format(order_id, order.state))
        order.cancel()
        self._order_closed(order)
        self._push_order_update(order)

    def subscribe_to_order_events(self, event_handler: OrderEventHandler):
        """
        Registers an async callable that will be called when an order is updated.

        :param event_handler: The event handler.
        """
        self._ctx.dispatcher.subscribe(self._order_updates.obj, cast(dispatcher.EventHandler, event_handler))

    def _update_balances(self, order: Order, balance_updates: ValueMapDict):
        # If we have holds associated with the order, it may be time to release some/all of those.
        hold_updates = {}
        order_holds = self._holds_by_order.get(order.id, ValueMap())
        if order_holds:
            if order.is_open:
                hold_updates = {
                    symbol: max(amount, -order_holds.get(symbol, Decimal(0)))
                    for symbol, amount in balance_updates.items()
                    if amount < Decimal(0) and symbol in order_holds
                }
            else:
                hold_updates = {symbol: -amount for symbol, amount in order_holds.items()}

        # Update holds and balances.
        if balance_updates or hold_updates:
            self._ctx.account_balances.update(balance_updates=balance_updates, hold_updates=hold_updates)

        # Update holds by order.
        if order_holds:
            if order.is_open:
                order_holds += hold_updates
            else:
                del self._holds_by_order[order.id]

    def _borrow(self, required_balances: ValueMap, order: Order):
        post_hold = {
            symbol: self._ctx.account_balances.get_available_balance(symbol) - required_amount
            for symbol, required_amount in required_balances.items()
        }
        balances_short = {
            symbol: -amount for symbol, amount in post_hold.items() if amount < Decimal(0)
        }

        loan_ids: List[str] = []
        try:
            # Create a loan for every balance that we're short on.
            for symbol, amount in balances_short.items():
                logger.debug(logs.StructuredMessage(
                    "Borrowing for order", order_id=order.id, symbol=symbol, amount=amount
                ))
                loan = self._ctx.loan_mgr.create_loan(symbol, amount)
                loan_ids.append(loan.id)
        except Exception as e:
            logger.debug(logs.StructuredMessage(
                "Failed to borrow", order_id=order.id, symbol=symbol, amount=amount, error=str(e)
            ))
            # If there are any errors borrowing money we'll rollback everything before propagating the exception.
            for loan_id in loan_ids:
                logger.debug(logs.StructuredMessage("Canceling loan", order_id=order.id, loan_id=loan_id))
                self._ctx.loan_mgr.cancel_loan(loan_id)
            raise

        # Add loans to order.
        for loan_id in loan_ids:
            order.add_loan(loan_id)

    def _repay_loans(self, order: Order):
        if order.operation == OrderOperation.BUY:
            credit_symbol = order.pair.base_symbol
        else:
            credit_symbol = order.pair.quote_symbol

        candidate_loans = [
            loan for loan in self._ctx.loan_mgr.get_loans(is_open=True)
            if loan.borrowed_symbol == credit_symbol
        ]
        # Try to cancel bigger loans first.
        candidate_loans.sort(key=lambda loan: loan.borrowed_amount, reverse=True)
        loan_ids: List[str] = []
        for loan in candidate_loans:
            try:
                self._ctx.loan_mgr.repay_loan(loan.id)
                loan_ids.append(loan.id)
                loan = cast(lending.LoanInfo, self._ctx.loan_mgr.get_loan(loan.id))
                logger.debug(logs.StructuredMessage("Repayed loan", loan=dataclasses.asdict(loan)))
            except errors.NotEnoughBalance:
                pass

        # Add loans to order.
        for loan_id in loan_ids:
            order.add_loan(loan_id)

    def _order_closed(self, order: Order):
        # The order is closed and there might be balances on hold that have to be released.
        self._update_balances(order, {})
        # If the order has auto_repay set and is filled, either fully or partially, then we need to cancel matching open
        # loans
        if order.auto_repay and order.amount_filled:
            self._repay_loans(order)

    def _process_order(
            self, order: Order, bar_event: bar.BarEvent, liquidity_strategy: liquidity.LiquidityStrategy
    ):
        def order_not_filled():
            order.not_filled()
            logger.debug(logs.StructuredMessage("Order not filled", order_id=order.id, order_state=order.state))
            if not order.is_open:
                self._order_closed(order)
                self._push_order_update(order, when=bar_event.when)

        # Calculate balance updates for the current bar.
        logger.debug(logs.StructuredMessage(
            "Processing order", order=order.get_debug_info(),
            bar={
                "open": bar_event.bar.open, "high": bar_event.bar.high, "low": bar_event.bar.low,
                "close": bar_event.bar.close, "volume": bar_event.bar.volume,
            }
        ))
        prev_state = order.state
        balance_updates = ValueMap(order.get_balance_updates(bar_event.bar, liquidity_strategy))
        assert order.state == prev_state, "The order state should not change inside get_balance_updates"
        self._round_balance_updates(balance_updates, order.pair)
        logger.debug(logs.StructuredMessage(
            "Order balance updates", order_id=order.id, balance_updates=balance_updates
        ))
        if order.pair.base_symbol not in balance_updates or order.pair.quote_symbol not in balance_updates:
            order_not_filled()
            return

        # Get fees, round them, and combine them with the balance updates.
        fees = ValueMap(self._ctx.fee_strategy.calculate_fees(order, balance_updates))
        self._round_fees(fees, order.pair)
        logger.debug(logs.StructuredMessage(
            "Order fees", order_id=order.id, fees=fees
        ))

        try:
            # Update balances. This may fail if there is not enough balance, so we do this first.
            final_updates = balance_updates + fees
            final_updates.prune()
            self._update_balances(order, final_updates)
            # Update the liquidity strategy.
            liquidity_strategy.take_liquidity(abs(balance_updates[bar_event.bar.pair.base_symbol]))
            # Update the order and release any pending balance on hold if the order is now closed.
            order.add_fill(bar_event.when, balance_updates, fees)
            logger.debug(logs.StructuredMessage(
                "Order updated", order_id=order.id, final_updates=final_updates, order_state=order.state
            ))

            if not order.is_open:
                self._order_closed(order)

            self._push_order_update(order, when=bar_event.when)

        except errors.NotEnoughBalance as e:
            logger.debug(logs.StructuredMessage(
                "Balance short processing order", order=order.get_debug_info(), error=str(e)
            ))
            order_not_filled()

    def _round_balance_updates(self, balance_updates: ValueMap, pair: Pair):
        pair_info = self._ctx.config.get_pair_info(pair)

        # For the base amount we truncate instead of rounding to avoid exceeding available liquidity.
        if (base_amount := balance_updates.get(pair.base_symbol)) is not None:
            balance_updates[pair.base_symbol] = core_helpers.truncate_decimal(base_amount, pair_info.base_precision)

        # For the quote amount we simply round.
        if (quote_amount := balance_updates.get(pair.quote_symbol)) is not None:
            balance_updates[pair.quote_symbol] = core_helpers.round_decimal(quote_amount, pair_info.quote_precision)

        balance_updates.prune()

    def _round_fees(self, fees: ValueMap, pair: Pair):
        pair_info = self._ctx.config.get_pair_info(pair)
        precisions = {
            pair.base_symbol: pair_info.base_precision,
            pair.quote_symbol: pair_info.quote_precision,
        }
        for symbol, precision in precisions.items():
            if (amount := fees.get(symbol)) is not None:
                fees[symbol] = core_helpers.round_decimal(amount, precision, rounding=decimal.ROUND_UP)

        fees.prune()

    def _estimate_required_balances(self, order: Order) -> ValueMap:
        # Get an estimated fill price. If the order can't provide one, use the last price available.
        estimated_fill_price = order.calculate_estimated_fill_price()
        if not estimated_fill_price:
            try:
                estimated_fill_price = self._ctx.prices.get_price(order.pair)
            except errors.NoPrice:
                pass

        # Build a dictionary of balance updates suitable for calculating fees.
        base_sign = helpers.get_base_sign_for_operation(order.operation)
        estimated_balance_updates = ValueMap({
            order.pair.base_symbol: order.amount * base_sign
        })
        if estimated_fill_price:
            estimated_balance_updates[order.pair.quote_symbol] = order.amount * estimated_fill_price * -base_sign
        self._round_balance_updates(estimated_balance_updates, order.pair)

        # Calculate fees.
        fees = ValueMap()
        if len(estimated_balance_updates) == 2:
            fees += self._ctx.fee_strategy.calculate_fees(order, estimated_balance_updates)
            self._round_fees(fees, order.pair)
        estimated_balance_updates += fees

        # Return only negative balance updates as required balances.
        return ValueMap({
            symbol: -amount for symbol, amount in estimated_balance_updates.items()
            if amount < Decimal(0)
        })

    def _push_order_update(self, order: Order, when: Optional[datetime.datetime] = None):
        # Checking dispatcher.now_available is necessary to avoid calling dispatcher.now() when no events have been
        # processed yet.
        if when is None and self._ctx.dispatcher.now_available:
            when = self._ctx.dispatcher.now()
        if when and self._order_updates.initialized:
            self._order_updates.push(OrderEvent(when, order.get_order_info()))



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Dict, List, Optional
import datetime

from . import client, common, config, helpers, spot_requests, user_data, websockets, websocket_mgr
from .client import spot as spot_client
from basana.core.config import get_config_value
from basana.core.enums import OrderOperation
from basana.core.pair import Pair


# Forward declarations
Balance = common.Balance
CanceledOCOOrder = common.CanceledOCOOrder
CanceledOrder = common.CanceledOrder
CreatedOCOOrder = common.CreatedOCOOrder
OCOOrderInfo = common.OCOOrderInfo
OCOOrderWrapper = common.OCOOrderWrapper
OrderEvent = user_data.OrderEvent
OrderEventHandler = user_data.OrderEventHandler
OrderInfo = common.OrderInfo
OrderUpdate = user_data.OrderUpdate
UserDataEvent = user_data.Event
UserDataEventHandler = user_data.UserDataEventHandler


class Trade(common.Trade):
    @property
    def order_list_id(self) -> Optional[str]:
        """The order list id."""
        ret = self.json.get("orderListId")
        ret = None if ret in [None, -1] else str(ret)
        return ret


class Fill(common.Fill):
    @property
    def trade_id(self) -> str:
        """The trade id."""
        return str(self.json["tradeId"])


class CreatedOrder(common.CreatedOrder):
    @property
    def order_list_id(self) -> Optional[str]:
        """The order list id."""
        ret = self.json["orderListId"]
        ret = None if ret == -1 else str(ret)
        return ret

    @property
    def fills(self) -> List[Fill]:
        # Only available for FULL responses.
        return [Fill(fill) for fill in self.json.get("fills", [])]


class OpenOrder(common.OpenOrder):
    @property
    def order_list_id(self) -> Optional[str]:
        """The order list id."""
        ret = self.json.get("orderListId")
        ret = None if ret in [None, -1] else str(ret)
        return ret

    @property
    def quote_amount(self) -> Optional[Decimal]:
        """The order amount in quote units."""
        return helpers.get_optional_decimal(self.json, "origQuoteOrderQty", True)


class SpotUserDataChannel(websockets.Channel):
    def __init__(self):
        self._listen_key = None

    @property
    def alias(self) -> str:
        return "spot_user_data"

    @property
    def stream(self) -> str:
        assert self._listen_key, "resolve_stream_name not called"
        return self._listen_key

    async def resolve_stream_name(self, api_client: client.APIClient):
        self._listen_key = (await api_client.spot_account.create_listen_key())["listenKey"]

    def keep_alive_period(self, config_overrides: dict = {}) -> Optional[datetime.timedelta]:
        return datetime.timedelta(
            seconds=get_config_value(
                config.DEFAULTS, "api.websockets.spot.user_data_stream.heartbeat", overrides=config_overrides
            )
        )

    async def keep_alive(self, api_client: client.APIClient):
        assert self._listen_key, "resolve_stream_name not called"
        await api_client.spot_account.keep_alive_listen_key(self._listen_key)


class Account:
    """Spot account."""
    def __init__(self, cli: spot_client.SpotAccount, ws_mgr: websocket_mgr.WebsocketManager):
        self._cli = cli
        self._ws_mgr = ws_mgr

    async def get_balances(self) -> Dict[str, Balance]:
        """Returns all balances."""
        account_info = await self._cli.get_account_information()
        return {balance["asset"].upper(): Balance(balance) for balance in account_info["balances"]}

    async def create_order(self, order_request: spot_requests.ExchangeOrder) -> CreatedOrder:
        created_order = await order_request.create_order(self._cli)
        return CreatedOrder(created_order)

    async def create_market_order(
            self, operation: OrderOperation, pair: Pair, amount: Optional[Decimal] = None,
            quote_amount: Optional[Decimal] = None, client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> CreatedOrder:
        """Creates a market order.

        Check https://binance-docs.github.io/apidocs/spot/en/#new-order-trade for more information.
        If the order can't be created a :class:`basana.external.binance.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell in base units.
        :param quote_amount: The amount to buy/sell in quote units.
        :param client_order_id: A client order id.
        :param kwargs: Additional keyword arguments that will be forwarded.

        .. note::

          * Either amount or quote_amount should be set, but not both.
        """

        return await self.create_order(spot_requests.MarketOrder(
            operation, pair, amount=amount, quote_amount=quote_amount, client_order_id=client_order_id, **kwargs
        ))

    async def create_limit_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal,
            time_in_force: str = "GTC", client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> CreatedOrder:
        """Creates a limit order.

        Check https://binance-docs.github.io/apidocs/spot/en/#new-order-trade for more information.
        If the order can't be created a :class:`basana.external.binance.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell in base units.
        :param limit_price: The limit price.
        :param time_in_force: The time in force.
        :param client_order_id: A client order id.
        :param kwargs: Additional keyword arguments that will be forwarded.
        """

        return await self.create_order(spot_requests.LimitOrder(
            operation, pair, amount, limit_price, time_in_force=time_in_force, client_order_id=client_order_id,
            **kwargs
        ))

    async def create_stop_limit_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, stop_price: Decimal, limit_price: Decimal,
            time_in_force: str = "GTC", client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> CreatedOrder:
        """Creates a stop limit order.

        Check https://binance-docs.github.io/apidocs/spot/en/#new-order-trade for more information.
        If the order can't be created a :class:`basana.external.binance.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell in base units.
        :param stop_price: The stop price.
        :param limit_price: The limit price.
        :param time_in_force: The time in force.
        :param client_order_id: A client order id.
        :param kwargs: Additional keyword arguments that will be forwarded.
        """

        return await self.create_order(spot_requests.StopLimitOrder(
            operation, pair, amount, stop_price, limit_price, time_in_force=time_in_force,
            client_order_id=client_order_id, **kwargs
        ))

    async def get_order_info(
            self, pair: Pair, order_id: Optional[str] = None, client_order_id: Optional[str] = None,
            include_trades: bool = True
    ) -> OrderInfo:
        """Returns information about an order.

        :param pair: The trading pair.
        :param order_id: The order id.
        :param client_order_id: The client order id.
        :param include_trades: True to include trades in the order info, False otherwise.

        .. note::

          * Either order_id or client_order_id should be set, but not both.
          * Including trades requires making an extra request to Binance.
        """
        order_book_symbol = helpers.pair_to_order_book_symbol(pair)
        order_info = await self._cli.query_order(
            order_book_symbol, order_id=None if order_id is None else int(order_id),
            orig_client_order_id=client_order_id
        )
        trades = []
        if include_trades:
            trades = [
                Trade(trade) for trade in await self._cli.get_trades(order_book_symbol, order_id=order_info["orderId"])
            ]
        return OrderInfo(order_info, trades)

    async def get_open_orders(self, pair: Optional[Pair] = None) -> List[OpenOrder]:
        """Returns open orders.

        :param pair: If set, only open orders matching this pair will be returned, otherwise all open orders will be
            returned.
        """

        order_book_symbol = None
        if pair:
            order_book_symbol = helpers.pair_to_order_book_symbol(pair)
        return [
            OpenOrder(open_order) for open_order in await self._cli.get_open_orders(order_book_symbol)
        ]

    async def cancel_order(
            self, pair: Pair, order_id: Optional[str] = None, client_order_id: Optional[str] = None,
    ) -> CanceledOrder:
        """Cancels an order.

        If the order can't be canceled a :class:`basana.external.binance.exchange.Error` will be raised.

        :param pair: The trading pair.
        :param order_id: The order id.
        :param client_order_id: The client order id.

        .. note::

          * Either order_id or client_order_id should be set, but not both.
        """
        canceled_order = await self._cli.cancel_order(
            helpers.pair_to_order_book_symbol(pair), order_id=None if order_id is None else int(order_id),
            orig_client_order_id=client_order_id
        )
        return CanceledOrder(canceled_order)

    async def create_oco_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal, stop_price: Decimal,
            stop_limit_price: Optional[Decimal] = None, stop_limit_time_in_force: str = "GTC",
            list_client_order_id: Optional[str] = None, limit_client_order_id: Optional[str] = None,
            stop_client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> CreatedOCOOrder:
        """Creates an OCO order.

        Check https://binance-docs.github.io/apidocs/spot/en/#new-oco-trade for more information.
        If the order can't be created a :class:`basana.external.binance.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell in base units.
        :param limit_price: The limit price.
        :param stop_price: The stop price.
        :param stop_limit_price: The stop limit price.
        :param stop_limit_time_in_force: The time in force for the stop limit order.
        :param list_client_order_id: A client id for the order list.
        :param limit_client_order_id: A client id for the limit order.
        :param stop_client_order_id: A client id for the stop order.
        :param kwargs: Additional keyword arguments that will be forwarded.
        """
        order_req = spot_requests.OCOOrder(
            operation, pair, amount, limit_price, stop_price, stop_limit_price=stop_limit_price,
            stop_limit_time_in_force=stop_limit_time_in_force, list_client_order_id=list_client_order_id,
            limit_client_order_id=limit_client_order_id, stop_client_order_id=stop_client_order_id,
            **kwargs
        )
        created_order = await order_req.create_order(self._cli)
        return CreatedOCOOrder(created_order)

    async def get_oco_order_info(
            self, order_list_id: Optional[str] = None, client_order_list_id: Optional[str] = None,
    ) -> OCOOrderInfo:
        """Returns information about an OCO order.

        :param order_list_id: The order list id.
        :param client_order_list_id: A client id for the order list.

        .. note::

          * Either order_list_id or client_order_list_id should be set, but not both.
        """
        order_info = await self._cli.query_oco_order(
            order_list_id=None if order_list_id is None else int(order_list_id),
            client_order_list_id=client_order_list_id
        )
        return OCOOrderInfo(order_info)

    async def cancel_oco_order(
            self, pair: Pair, order_list_id: Optional[str] = None, client_order_list_id: Optional[str] = None,
    ) -> CanceledOCOOrder:
        """Cancels an OCO order.

        If the order can't be canceled a :class:`basana.external.binance.exchange.Error` will be raised.

        :param pair: The trading pair.
        :param order_list_id: The order list id.
        :param client_order_list_id: A client id for the order list.

        .. note::

          * Either order_list_id or client_order_list_id should be set, but not both.
        """

        canceled_order = await self._cli.cancel_oco_order(
            helpers.pair_to_order_book_symbol(pair),
            order_list_id=None if order_list_id is None else int(order_list_id),
            client_order_list_id=client_order_list_id
        )
        return CanceledOCOOrder(canceled_order)

    def subscribe_to_user_data_events(self, event_handler: UserDataEventHandler):
        """
        Registers an async callable that will be called for every new user data event.

        Works as defined in https://developers.binance.com/docs/binance-spot-api-docs/user-data-stream.

        :param event_handler: The event handler.
        """

        self._ws_mgr.subscribe_to_user_data_events(
            SpotUserDataChannel(),
            lambda ws_cli: user_data.WebSocketEventSource(ws_cli),
            event_handler
        )

    def subscribe_to_order_events(self, event_handler: OrderEventHandler):
        """
        Registers an async callable that will be called for every new order update.

        Works as defined in https://developers.binance.com/docs/binance-spot-api-docs/user-data-stream#order-update.

        :param event_handler: The event handler.
        """

        self._ws_mgr.subscribe_to_order_events(
            SpotUserDataChannel(),
            lambda ws_cli: user_data.WebSocketEventSource(ws_cli),
            event_handler
        )



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Awaitable, Callable, Dict, Optional
import datetime

from . import helpers
from basana.core import event, websockets as core_ws
from basana.core.enums import OrderOperation


class OrderUpdate:
    def __init__(self, json: dict):
        assert json["e"] == "executionReport"

        #: The JSON representation.
        self.json: dict = json

    @property
    def id(self) -> str:
        """The order id."""
        return str(self.json["i"])

    @property
    def symbol(self) -> str:
        """Symbol."""
        return str(self.json["s"])

    @property
    def client_order_id(self) -> str:
        """The client order id."""
        return self.json["c"]

    @property
    def operation(self) -> OrderOperation:
        """The operation."""
        return helpers.side_to_order_operation(self.json["S"])

    @property
    def type(self) -> str:
        """
        The type of order.

        Check **Order types** in
        https://developers.binance.com/docs/binance-spot-api-docs/enums#order-types-ordertypes-type.
        """
        return self.json["o"]

    @property
    def time_in_force(self) -> Optional[str]:
        """
        The time in force.

        Check **Time in force** in
        https://developers.binance.com/docs/binance-spot-api-docs/enums#time-in-force-timeinforce.
        """
        return self.json.get("f")

    @property
    def amount(self) -> Decimal:
        """The amount."""
        return Decimal(self.json["q"])

    @property
    def quote_amount(self) -> Optional[Decimal]:
        """The amount set in quote units."""
        return helpers.get_optional_decimal(self.json, "Q", True)

    @property
    def limit_price(self) -> Optional[Decimal]:
        """The limit price."""
        return helpers.get_optional_decimal(self.json, "p", True)

    @property
    def stop_price(self) -> Optional[Decimal]:
        """The stop price."""
        return helpers.get_optional_decimal(self.json, "P", True)

    @property
    def order_list_id(self) -> Optional[str]:
        """The order list id."""
        ret = self.json.get("g")
        ret = None if ret in [None, -1] else str(ret)
        return ret

    @property
    def status(self) -> str:
        """
        The status.

        Check **Order status** in https://developers.binance.com/docs/binance-spot-api-docs/enums#order-status-status.
        """
        return self.json["X"]

    @property
    def is_open(self) -> bool:
        """True if the order is open, False otherwise."""
        return helpers.order_status_is_open(self.status)

    @property
    def amount_filled(self) -> Decimal:
        """The cumulative amount filled."""
        return Decimal(self.json["z"])

    @property
    def quote_amount_filled(self) -> Decimal:
        """The cumulative amount filled in quote units."""
        return Decimal(self.json["Z"])

    @property
    def fees(self) -> Dict[str, Decimal]:
        """The fees."""
        ret = {}
        if commision_asset := self.json.get("N"):
            ret[commision_asset] = Decimal(self.json["n"])
        return ret


class Event(event.Event):
    def __init__(self, when: datetime.datetime, json: dict):
        super().__init__(when)
        #: The JSON representation.
        self.json: dict = json


class OrderEvent(Event):
    """
    An event for order updates.

    :param when: The datetime when the event occurred. It must have timezone information set.
    :param order_update: The order update.
    """

    def __init__(self, when: datetime.datetime, json: dict):
        super().__init__(when, json)

        #: The order update.
        self.order_update: OrderUpdate = OrderUpdate(json)


# Generate events from websocket messages.
class WebSocketEventSource(core_ws.ChannelEventSource):
    def __init__(self, producer: event.Producer):
        super().__init__(producer=producer)

    async def push_from_message(self, message: dict):
        json = message["data"]
        # Push the event.
        event_cls = {
            "executionReport": OrderEvent,
        }.get(json["e"], Event)
        event = event_cls(helpers.timestamp_to_datetime(int(json["E"])), json)
        self.push(event)


OrderEventHandler = Callable[[OrderEvent], Awaitable[Any]]
UserDataEventHandler = Callable[[Event], Awaitable[Any]]



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import cast, Any, Awaitable, Callable, Dict, List, Optional, Tuple, Union
import datetime
import enum

from dateutil.parser import parse as dt_parse
import aiohttp

from . import client, helpers, requests, order_book, orders, trades, websockets as bitstamp_ws
from basana.core import bar, dispatcher, event, token_bucket, websockets as core_ws
from basana.core.enums import OrderOperation
from basana.core.pair import Pair, PairInfo


BarEventHandler = Callable[[bar.BarEvent], Awaitable[Any]]
Error = client.Error
OrderBookEvent = order_book.OrderBookEvent
OrderBookEventHandler = Callable[[order_book.OrderBookEvent], Awaitable[Any]]
OrderEvent = orders.OrderEvent
OrderEventHandler = Callable[[orders.OrderEvent], Awaitable[Any]]
TradeEvent = trades.TradeEvent
TradeEventHandler = Callable[[trades.TradeEvent], Awaitable[Any]]


@enum.unique
class TransactionType(enum.IntEnum):
    """Enumeration for transaction types."""

    #:
    DEPOSIT = 0

    #:
    WITHDRAWAL = 1

    #:
    MARKET_TRADE = 2


class OpenOrder:
    def __init__(self, json: dict):
        self.json = json

    @property
    def id(self) -> str:
        """The order id."""
        return str(self.json["id"])

    @property
    def datetime(self) -> datetime.datetime:
        """The creation datetime."""
        return dt_parse(self.json["datetime"]).replace(tzinfo=datetime.timezone.utc)

    @property
    def operation(self) -> OrderOperation:
        """The operation."""
        return helpers.order_type_to_order_operation(int(self.json["type"]))

    @property
    def limit_price(self) -> Decimal:
        """The limit price."""
        return Decimal(self.json["price"])

    @property
    def amount(self) -> Decimal:
        """The amount."""
        return Decimal(self.json["amount_at_create"])

    @property
    def amount_filled(self) -> Decimal:
        """The amount filled."""
        return Decimal(self.json["amount"])

    @property
    def pair(self) -> Pair:
        """The trading pair."""
        return Pair(*self.json["currency_pair"].split("/"))

    @property
    def client_order_id(self) -> Optional[str]:
        """The client order id."""
        return self.json.get("client_order_id")


class OrderStatusTransaction:
    def __init__(self, json: dict):
        self.json = json

    @property
    def tid(self) -> str:
        """The transaction id."""
        return str(self.json["tid"])

    @property
    def price(self) -> Decimal:
        """The price."""
        return Decimal(self.json["price"])

    @property
    def fee(self) -> Decimal:
        """The fee."""
        return Decimal(self.json["fee"])

    @property
    def type(self) -> TransactionType:
        """The transaction type."""
        return TransactionType(int(self.json["type"]))

    def __getattr__(self, item) -> Decimal:
        """Use for numeric, dynamic attributes, like btc, usd, usdt, etc."""
        ret = self.json.get(item)
        if ret is None:
            raise AttributeError("f{item} not found")
        return Decimal(ret)


class OrderStatus:
    def __init__(self, json: dict):
        self.json = json

    @property
    def id(self) -> str:
        """The order id."""
        return str(self.json["id"])

    @property
    def status(self) -> str:
        """The status. One of Open, Finished, Expired or Canceled"""
        return self.json["status"]

    @property
    def amount_remaining(self) -> Decimal:
        """The amount remaining to be filled."""
        return Decimal(self.json["amount_remaining"])

    @property
    def client_order_id(self) -> Optional[str]:
        """The client order id."""
        return self.json.get("client_order_id")

    @property
    def transactions(self) -> List[OrderStatusTransaction]:
        """The transactions.

        Check https://www.bitstamp.net/api/#order-status.
        """
        return [OrderStatusTransaction(tx) for tx in self.json.get("transactions", [])]


class OrderInfo:
    def __init__(self, pair: Pair, order_status: OrderStatus):
        self._pair = pair
        self._order_status = order_status

        fees = Decimal(0)
        self._filled_base_amount = Decimal(0)
        self._quote_amount_filled = Decimal(0)
        base_currency = pair.base_symbol.lower()
        quote_currency = pair.quote_symbol.lower()
        for tx in order_status.transactions:
            fees += tx.fee
            self._filled_base_amount += getattr(tx, base_currency)
            self._quote_amount_filled += getattr(tx, quote_currency)
        self._fees = {pair.quote_symbol: fees} if fees else {}

    @property
    def id(self) -> str:
        """The order id."""
        return self._order_status.id

    @property
    def is_open(self) -> bool:
        """True if the order is open, False otherwise."""
        # Not doing `order_status.status == "Open"` to detect unkown states.
        is_open = {
            "Open": True,
            "Finished": False,
            "Expired": False,
            "Canceled": False,
        }.get(self._order_status.status)
        assert is_open is not None, "No mapping for {} order status".format(
            self._order_status.status
        )
        return is_open

    @property
    def amount_filled(self) -> Decimal:
        """The amount filled."""
        return self._filled_base_amount

    @property
    def amount_remaining(self) -> Decimal:
        """The amount remaining to be filled."""
        return self._order_status.amount_remaining

    @property
    def quote_amount_filled(self) -> Decimal:
        """The amount filled in quote units."""
        return self._quote_amount_filled

    @property
    def fill_price(self) -> Optional[Decimal]:
        """The fill price."""
        fill_price = None
        if self.amount_filled:
            fill_price = self.quote_amount_filled / self.amount_filled
        return fill_price

    @property
    def fees(self) -> Dict[str, Decimal]:
        """The fees."""
        return self._fees


class Balance:
    def __init__(self, json: dict):
        self.json = json

    @property
    def available(self) -> Decimal:
        """The available balance."""
        return Decimal(self.json["available"])

    @property
    def total(self) -> Decimal:
        """The total balance (available + reserved)."""
        return Decimal(self.json["total"])

    @property
    def reserved(self) -> Decimal:
        """The reserved balance."""
        return Decimal(self.json["reserved"])


class CanceledOrder:
    def __init__(self, json: dict):
        self.json = json

    @property
    def id(self) -> str:
        """The order id."""
        return str(self.json["id"])

    @property
    def amount(self) -> Decimal:
        """The amount."""
        return Decimal(self.json["amount"])

    @property
    def limit_price(self) -> Decimal:
        """The limit price."""
        return Decimal(self.json["price"])

    @property
    def operation(self) -> OrderOperation:
        """The operation."""
        return helpers.order_type_to_order_operation(int(self.json["type"]))


class CreatedOrder:
    def __init__(self, json: dict):
        self.json = json

    @property
    def id(self) -> str:
        """The order id."""
        return str(self.json["id"])

    @property
    def datetime(self) -> datetime.datetime:
        """The creation datetime."""
        return dt_parse(self.json["datetime"]).replace(tzinfo=datetime.timezone.utc)

    @property
    def operation(self) -> OrderOperation:
        """The operation."""
        return helpers.order_type_to_order_operation(int(self.json["type"]))

    @property
    def price(self) -> Decimal:
        """The price."""
        return Decimal(self.json["price"])

    @property
    def amount(self) -> Decimal:
        """The amount."""
        return Decimal(self.json["amount"])

    @property
    def client_order_id(self) -> Optional[str]:
        """The client order id."""
        return self.json.get("client_order_id")


class RealTimeTradesToBar(bar.RealTimeTradesToBar):
    async def on_trade_event(self, trade_event: trades.TradeEvent):
        self.push_trade(trade_event.trade.datetime, trade_event.trade.price, trade_event.trade.amount)


class Exchange:
    """A client for `Bitstamp <https://www.bitstamp.net//>`_ crypto currency exchange.

    :param dispatcher: The event dispatcher.
    :param api_key: An optional api key. If not set only public endpoints can be used.
    :param api_secret: An optional api secret. If not set only public endpoints can be used.
    :param session: An optional client session, in case you want to reuse connections.
    :type session: aiohttp.ClientSession
    :param tb: An optional token bucket limiter, in case you want to throttle requests.
    :param config_overrides: An optional dictionary for overriding config settings.
    """
    def __init__(
            self, dispatcher: dispatcher.EventDispatcher, api_key: Optional[str] = None,
            api_secret: Optional[str] = None, session: Optional[aiohttp.ClientSession] = None,
            tb: Optional[token_bucket.TokenBucketLimiter] = None, config_overrides: dict = {}
    ):
        self._dispatcher = dispatcher
        self._api_key = api_key
        self._api_secret = api_secret
        self._cli = client.APIClient(api_key, api_secret, session=session, tb=tb, config_overrides=config_overrides)
        self._session = session
        self._tb = tb
        self._config_overrides = config_overrides
        self._pub_websocket: Optional[bitstamp_ws.PublicWebSocketClient] = None
        self._priv_websocket: Optional[bitstamp_ws.PrivateWebSocketClient] = None
        self._channel_to_event_source: Dict[str, event.EventSource] = {}
        self._pair_info_cache: Dict[Pair, PairInfo] = {}
        self._bar_event_source: Dict[Tuple[Pair, int, bool, float], RealTimeTradesToBar] = {}

    async def get_balance(self, symbol: str) -> Balance:
        """Returns the balance for a specific currency/symbol/etc..

        :param symbol: The currency/symbol/etc..
        """
        balance = await self._cli.get_account_balance(symbol.lower())
        return Balance(balance)

    async def get_bid_ask(self, pair: Pair) -> Tuple[Decimal, Decimal]:
        """Returns the current bid and ask price.

        :param pair: The trading pair.
        """
        ticker = await self._cli.get_ticker(helpers.pair_to_currency_pair(pair))
        return Decimal(ticker["bid"]), Decimal(ticker["ask"])

    async def fetch_pair_info(self):
        # Fetch trading pairs and fill the cache.
        bs_pairs_info = await self._cli.get_trading_pairs_info()
        for bs_pair_info in bs_pairs_info:
            pair = Pair(*bs_pair_info["name"].split("/"))
            pair_info = PairInfo(int(bs_pair_info["base_decimals"]), int(bs_pair_info["counter_decimals"]))
            self._pair_info_cache[pair] = pair_info

    async def get_pair_info(self, pair: Pair) -> PairInfo:
        """Returns information about a trading pair.

        :param pair: The trading pair.
        """
        if not self._pair_info_cache:
            await self.fetch_pair_info()
        return self._pair_info_cache[pair]

    async def create_order(self, order_request: requests.ExchangeOrder) -> CreatedOrder:
        created_order = await order_request.create_order(self._cli)
        return CreatedOrder(created_order)

    async def create_market_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, client_order_id: Optional[str] = None,
            **kwargs: Dict[str, Any]
    ) -> CreatedOrder:
        """Creates a market order.

        If the order can't be created a :class:`basana.external.bitstamp.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell in base units.
        :param client_order_id: A client order id.
        :param kwargs: Additional keyword arguments that will be forwarded.
        """
        return await self.create_order(requests.MarketOrder(
            operation, pair, amount, client_order_id=client_order_id, **kwargs
        ))

    async def create_limit_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal,
            client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> CreatedOrder:
        """Creates a limit order.

        If the order can't be created a :class:`basana.external.bitstamp.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell in base units.
        :param limit_price: The limit price.
        :param client_order_id: A client order id.
        :param kwargs: Additional keyword arguments that will be forwarded.
        """
        return await self.create_order(requests.LimitOrder(
            operation, pair, amount, limit_price, client_order_id=client_order_id, **kwargs
        ))

    async def create_instant_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, amount_in_counter: bool = False,
            client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> CreatedOrder:
        """Creates an instant order.

        If the order can't be created a :class:`basana.external.bitstamp.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell.
        :param amount_in_counter: False if the amount is set in base currency (default), True if the amount is set in
                                  quote currency.
        :param client_order_id: A client order id.
        :param kwargs: Additional keyword arguments that will be forwarded.
        """
        return await self.create_order(requests.InstantOrder(
            operation, pair, amount, amount_in_counter=amount_in_counter, client_order_id=client_order_id, **kwargs
        ))

    async def cancel_order(self, order_id: Union[str, int]) -> CanceledOrder:
        """Cancels an order.

        If the order can't be canceled a :class:`basana.external.bitstamp.exchange.Error` will be raised.

        :param order_id: The order id.
        """
        canceled_order = await self._cli.cancel_order(int(order_id))
        return CanceledOrder(canceled_order)

    async def get_order_info(
            self, pair: Pair, order_id: Optional[Union[str, int]] = None, client_order_id: Optional[str] = None,
    ) -> Optional[OrderInfo]:
        """Returns information about an order.

        :param pair: The trading pair.
        :param order_id: The order id.
        :param client_order_id: The client order id.

        .. note::

          * Either order_id or client_order_id should be set, but not both.
          * For closed orders, this call only returns information for the last 30 days.
        """
        ret = None
        order_status = await self.get_order_status(order_id=order_id, client_order_id=client_order_id)
        if order_status:
            ret = OrderInfo(pair, order_status)
        return ret

    async def get_order_status(
            self, order_id: Optional[Union[str, int]] = None, client_order_id: Optional[str] = None,
            omit_transactions: Optional[bool] = None
    ) -> Optional[OrderStatus]:
        if order_id is not None:
            order_id = int(order_id)

        ret = None
        try:
            order_status = await self._cli.get_order_status(
                id=order_id, client_order_id=client_order_id, omit_transactions=omit_transactions
            )
            ret = OrderStatus(order_status)
        except client.Error as e:
            if e.reason != "Order not found.":
                raise
        return ret

    async def get_open_orders(self, pair: Optional[Pair] = None) -> List[OpenOrder]:
        """Returns open orders.

        :param pair: If set, only open orders matching this pair will be returned, otherwise all open orders will be
            returned.
        """
        currency_pair = None if pair is None else helpers.pair_to_currency_pair(pair)
        open_orders = await self._cli.get_open_orders(currency_pair=currency_pair)
        return [OpenOrder(open_order) for open_order in open_orders]

    async def get_balances(self) -> Dict[str, Balance]:
        """Returns all balances."""
        balances = await self._cli.get_account_balances()
        return {balance["currency"].upper(): Balance(balance) for balance in balances}

    def subscribe_to_bar_events(
            self, pair: Pair, bar_duration: int, event_handler: BarEventHandler, skip_first_bar: bool = True,
            flush_delay: float = 1
    ):
        """Registers an async callable that will be called when a new bar is available.


        :param pair: The trading pair.
        :param bar_duration: The bar duration in seconds.
        :param event_handler: An async callable that receives BarEvent.
        :param skip_first_bar: True if the first bar should be skipped. This is to avoid receiving incomplete bars
                               if subscription takes place in the middle of the period.
        :param flush_delay: The number of seconds to wait before generating the new bar, in case trades are delayed.

        .. note::

          * Under the hood, this will subscribe to trade events and will aggregate those into bars.
        """
        key = (pair, bar_duration, skip_first_bar, flush_delay)
        event_source = self._bar_event_source.get(key)
        if not event_source:
            event_source = RealTimeTradesToBar(
                pair, bar_duration, skip_first_bar=skip_first_bar, flush_delay=flush_delay
            )
            self.subscribe_to_public_trade_events(pair, event_source.on_trade_event)
            self._bar_event_source[key] = event_source
        self._dispatcher.subscribe(event_source, cast(dispatcher.EventHandler, event_handler))

    def subscribe_to_order_book_events(self, pair: Pair, event_handler: OrderBookEventHandler):
        """Registers an async callable that will be called when the order book is updated.

        :param pair: The trading pair.
        :param event_handler: An async callable that receives an OrderBookEvent.

        .. note::

          * Only the top 100 bids and asks are returned in the event.
        """
        channel = order_book.get_channel(pair)
        self._subscribe_to_ws_channel_events(
            channel, True, lambda ws_cli: order_book.WebSocketEventSource(pair, ws_cli),
            cast(dispatcher.EventHandler, event_handler)
        )

    def subscribe_to_public_order_events(self, pair: Pair, event_handler: OrderEventHandler):
        """Registers an async callable that will be called when any order is updated.

        :param pair: The trading pair.
        :param event_handler: An async callable that receives an OrderEvent.
        """
        channel = orders.get_public_channel(pair)
        self._subscribe_to_ws_channel_events(
            channel, True, lambda ws_cli: orders.WebSocketEventSource(pair, ws_cli),
            cast(dispatcher.EventHandler, event_handler)
        )

    def subscribe_to_private_order_events(self, pair: Pair, event_handler: OrderEventHandler):
        """Registers an async callable that will be called when your own orders are updated.

        :param pair: The trading pair.
        :param event_handler: An async callable that receives an OrderEvent.
        """
        channel = orders.get_private_channel(pair)
        self._subscribe_to_ws_channel_events(
            channel, False, lambda ws_cli: orders.WebSocketEventSource(pair, ws_cli),
            cast(dispatcher.EventHandler, event_handler)
        )

    def subscribe_to_public_trade_events(self, pair: Pair, event_handler: TradeEventHandler):
        """Registers an async callable that will be called for any new trade.

        :param pair: The trading pair.
        :param event_handler: An async callable that receives a TradeEvent.
        """
        channel = trades.get_public_channel(pair)
        self._subscribe_to_ws_channel_events(
            channel, True, lambda ws_cli: trades.WebSocketEventSource(pair, ws_cli),
            cast(dispatcher.EventHandler, event_handler)
        )

    def subscribe_to_private_trade_events(self, pair: Pair, event_handler: TradeEventHandler):
        """Registers an async callable that will be called for your own new trades.

        :param pair: The trading pair.
        :param event_handler: An async callable that receives a TradeEvent.
        """
        channel = trades.get_private_channel(pair)
        self._subscribe_to_ws_channel_events(
            channel, False, lambda ws_cli: trades.WebSocketEventSource(pair, ws_cli),
            cast(dispatcher.EventHandler, event_handler)
        )

    def _subscribe_to_ws_channel_events(
            self, channel: str, is_public: bool,
            event_src_factory: Callable[[core_ws.WebSocketClient], core_ws.ChannelEventSource],
            event_handler: dispatcher.EventHandler
    ):
        # Get/create the event source for the channel.
        ws_cli = self._get_pub_ws_client() if is_public else self._get_priv_ws_client()
        event_source = ws_cli.get_channel_event_source(channel)
        if not event_source:
            event_source = event_src_factory(ws_cli)
            ws_cli.set_channel_event_source(channel, event_source)

        # Subscribe the event handler to the event source.
        self._dispatcher.subscribe(event_source, event_handler)

    def _get_pub_ws_client(self) -> bitstamp_ws.WebSocketClient:
        if self._pub_websocket is None:
            self._pub_websocket = bitstamp_ws.PublicWebSocketClient(
                session=self._session, config_overrides=self._config_overrides
            )
        return self._pub_websocket

    def _get_priv_ws_client(self) -> bitstamp_ws.WebSocketClient:
        assert self._api_key and self._api_secret, "Missing API key and/or secret"

        if self._priv_websocket is None:
            self._priv_websocket = bitstamp_ws.PrivateWebSocketClient(
                self._api_key, self._api_secret, session=self._session, config_overrides=self._config_overrides
            )
        return self._priv_websocket



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Sequence
import abc
import codecs
import contextlib
import csv

from basana.core import event

# This module is not using async io. Why ?
# asyncio does not support asynchronous operations on the filesystem.
# Check https://github.com/python/asyncio/wiki/ThirdParty#filesystem.
# aiofiles could be used, but given that:
# * aiofiles delegates operations to a separate thread pool
# * this module will be used mostly for backtesting where there is no other IO taking place
# I decided not to support async io initially.


@contextlib.contextmanager
def open_file_with_detected_encoding(filename, default_encoding='utf-8'):
    with open(filename, 'rb') as file:
        raw = file.read(4)  # Read enough bytes to detect BOMs

    boms = [
        (codecs.BOM_UTF32_LE, 'utf-32-le'),
        (codecs.BOM_UTF32_BE, 'utf-32-be'),
        (codecs.BOM_UTF16_LE, "utf-16-le"),
        (codecs.BOM_UTF16_BE, "utf-16-be"),
        (codecs.BOM_UTF8, "utf-8-sig"),
    ]
    encoding = default_encoding
    offset = 0
    for bom, enc in boms:
        if raw.startswith(bom):
            encoding = enc
            offset = len(bom)
            break

    # Re-open the file with the detected encoding and skip the bom.
    f = open(filename, 'r', encoding=encoding)
    if offset:
        f.seek(offset)
    yield f


class RowParser(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def parse_row(self, row_dict: dict) -> Sequence[event.Event]:
        raise NotImplementedError()


def load_sort_and_yield(csv_path: str, row_parser: RowParser, dict_reader_kwargs: dict = {}):
    events = []

    # Load events.
    with open_file_with_detected_encoding(csv_path) as f:
        dict_reader = csv.DictReader(f, **dict_reader_kwargs)
        for row in dict_reader:
            for ev in row_parser.parse_row(row):
                events.append(ev)

    # Sort them for proper delivery.
    events = sorted(events, key=lambda ev: ev.when)

    for ev in events:
        yield ev


def load_and_yield(csv_path: str, row_parser: RowParser, dict_reader_kwargs: dict = {}):

    # Load events.
    with open_file_with_detected_encoding(csv_path) as f:
        dict_reader = csv.DictReader(f, **dict_reader_kwargs)
        for row in dict_reader:
            for ev in row_parser.parse_row(row):
                yield ev


class EventSource(event.EventSource, event.Producer):
    def __init__(self, csv_path: str, row_parser: RowParser, sort: bool = True, dict_reader_kwargs: dict = {}):
        super().__init__(producer=self)
        self._csv_path = csv_path
        self._row_parser = row_parser
        self._sort = sort
        self._dict_reader_kwargs = dict_reader_kwargs
        self._row_it = None

    async def initialize(self):
        if self._sort:
            self._row_it = load_sort_and_yield(self._csv_path, self._row_parser, self._dict_reader_kwargs)
        else:
            self._row_it = load_and_yield(self._csv_path, self._row_parser, self._dict_reader_kwargs)

    async def finalize(self):
        self._row_it = None

    def pop(self) -> Optional[event.Event]:
        ret = None
        try:
            if self._row_it:
                ret = next(self._row_it)
        except StopIteration:
            self._row_it = None
        return ret



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from basana.core import pair
from basana.core.event_sources import csv
from basana.external.binance.tools.download_bars import period_to_step
from basana.external.common.csv.bars import RowParser


period_to_timedelta = {
    period_str: datetime.timedelta(seconds=period_secs)
    for period_str, period_secs in period_to_step.items()
}


class BarSource(csv.EventSource):
    def __init__(
            self, pair: pair.Pair, csv_path: str, period: str,
            sort: bool = False, tzinfo: datetime.tzinfo = datetime.timezone.utc,
            dict_reader_kwargs: dict = {}
    ):
        # The datetime in the files are the beginning of the period but we need to generate the event at the period's
        # end.
        timedelta = period_to_timedelta.get(period)
        assert timedelta is not None, "Invalid period"
        self.row_parser = RowParser(pair, tzinfo=tzinfo, timedelta=timedelta)
        super().__init__(csv_path, self.row_parser, sort=sort, dict_reader_kwargs=dict_reader_kwargs)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
import datetime

from . import helpers
from basana.core import dt, event, websockets as core_ws
from basana.core.enums import OrderOperation
from basana.core.pair import Pair


class Trade:
    def __init__(self, pair: Pair, json: dict):
        #: The trading pair.
        self.pair: Pair = pair
        #: The JSON representation.
        self.json: dict = json

    @property
    def id(self) -> str:
        """The trade id."""
        return str(self.json["id"])

    @property
    def datetime(self) -> datetime.datetime:
        """The datetime when the trade occurred."""
        timestamp = int(self.json["microtimestamp"]) / 1e6
        return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)

    @property
    def amount(self) -> Decimal:
        """The amount."""
        return Decimal(self.json["amount_str"])

    @property
    def price(self) -> Decimal:
        """The price."""
        return Decimal(self.json["price_str"])

    @property
    def type(self) -> OrderOperation:
        # TODO: Deprecate this property.
        return self.operation

    @property
    def operation(self) -> OrderOperation:
        """The operation."""
        return helpers.order_type_to_order_operation(int(self.json["type"]))

    @property
    def buy_order_id(self) -> str:
        """The buy order id."""
        return str(self.json["buy_order_id"])

    @property
    def sell_order_id(self) -> str:
        """The sell order id."""
        return str(self.json["sell_order_id"])


class TradeEvent(event.Event):
    def __init__(self, when: datetime.datetime, trade: Trade):
        super().__init__(when)
        #: The trade.
        self.trade: Trade = trade


# Generate Trade events from websocket messages.
class WebSocketEventSource(core_ws.ChannelEventSource):
    def __init__(self, pair: Pair, producer: event.Producer):
        super().__init__(producer=producer)
        self._pair: Pair = pair

    async def push_from_message(self, message: dict):
        self.push(TradeEvent(dt.utc_now(), Trade(self._pair, message["data"])))


def get_public_channel(pair: Pair) -> str:
    return "live_trades_{}".format(helpers.pair_to_currency_pair(pair))


def get_private_channel(pair: Pair) -> str:
    return "private-my_trades_{}".format(helpers.pair_to_currency_pair(pair))



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
import abc

from basana.backtesting import errors
from basana.core import bar


class LiquidityStrategy(metaclass=abc.ABCMeta):
    """Base class for liquidity strategies.

    A liquidity strategy defines how much of a :class:`basana.Bar`s volume can be used when processing an order and
    what is the price slippage.

    .. note::

        * This is a base class and should not be used directly.
        * Concrete strategies will be created by the :class:`basana.backtesting.exchange.Exchange` for each traded
          pair.
    """

    @abc.abstractmethod
    def on_bar(self, bar: bar.Bar):
        """Called when a new bar is available."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def available_liquidity(self) -> Decimal:
        """Returns the available liquidity."""
        raise NotImplementedError()

    @abc.abstractmethod
    def take_liquidity(self, amount: Decimal) -> Decimal:
        """Takes/consumes available liquidity.

        :param amount: The amount of liquidity to take. It must be <= available liquidity.
        :returns: The percentage of the price impact.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def calculate_price_impact(self, amount: Decimal) -> Decimal:
        """A read-only version of :meth:`take_liquidity`."""
        raise NotImplementedError()

    @abc.abstractmethod
    def calculate_amount(self, price_impact: Decimal) -> Decimal:
        """Returns the amount of liquidity that can be taken with an impact <= price_impact."""
        raise NotImplementedError()


class InfiniteLiquidity(LiquidityStrategy):
    """This class models infinite liquidity with no price impact."""

    def on_bar(self, bar: bar.Bar):
        pass

    @property
    def available_liquidity(self) -> Decimal:
        return Decimal("Infinity")

    def take_liquidity(self, amount: Decimal) -> Decimal:
        assert amount > Decimal(0)
        assert amount <= self.available_liquidity

        return Decimal(0)

    def calculate_price_impact(self, amount: Decimal) -> Decimal:
        assert amount > Decimal(0)
        assert amount <= self.available_liquidity

        return Decimal(0)

    def calculate_amount(self, price_impact: Decimal) -> Decimal:
        assert price_impact >= Decimal(0), f"Invalid price_impact {price_impact}"

        return Decimal("Infinity")


class VolumeShareImpact(LiquidityStrategy):
    """
    The price impact is calculated by multiplying the price impact constant by the square of the ratio of the used
    volume to the total volume.

    :param volume_limit_pct: Maximum percentage of volume that can be used from each bar.
    :param price_impact: Maximum price impact (percentage).
    """

    def __init__(self, volume_limit_pct: Decimal = Decimal("25"), price_impact: Decimal = Decimal("10")):
        assert volume_limit_pct >= Decimal(0), f"Invalid volume_limit_pct {volume_limit_pct}"
        assert price_impact >= Decimal(0), f"Invalid price_impact {price_impact}"

        self._volume_limit_pct = volume_limit_pct / Decimal(100)
        self._price_impact_pct = price_impact / Decimal(100)
        self._total_liquidity = Decimal(0)
        self._used_liquidity = Decimal(0)

    def on_bar(self, bar: bar.Bar):
        self._total_liquidity = bar.volume * self._volume_limit_pct
        self._used_liquidity = Decimal(0)

    def _volume_share_impact(self, used_liquidity: Decimal) -> Decimal:
        # impact = (used_liquidity / (used_liquidity + available_liquidity)) ** 2 * price_impact
        assert used_liquidity >= Decimal(0), f"Invalid used_liquidity {used_liquidity}"
        assert used_liquidity <= self._total_liquidity, f"used_liquidity {used_liquidity} too high"

        if used_liquidity == Decimal(0):
            ret = Decimal(0)
        else:
            used_pct = used_liquidity / self._total_liquidity
            ret = used_pct ** Decimal(2) * self._price_impact_pct
        return ret

    @property
    def available_liquidity(self) -> Decimal:
        return self._total_liquidity - self._used_liquidity

    def take_liquidity(self, amount: Decimal) -> Decimal:
        assert amount >= Decimal(0), f"Invalid amount {amount}"
        if amount > self.available_liquidity:
            raise errors.Error("Not enough liquidity")

        impact_pre = self._volume_share_impact(self._used_liquidity)
        self._used_liquidity += amount
        impact_post = self._volume_share_impact(self._used_liquidity)
        diff = impact_post - impact_pre
        assert diff >= Decimal(0)
        return diff

    def calculate_price_impact(self, amount: Decimal) -> Decimal:
        assert amount >= Decimal(0), f"Invalid amount {amount}"
        if amount > self.available_liquidity:
            raise errors.Error("Not enough liquidity")

        return self._volume_share_impact(self._used_liquidity + amount)

    def calculate_amount(self, price_impact: Decimal) -> Decimal:
        assert price_impact >= Decimal(0), f"Invalid price_impact {price_impact}"

        # price_impact = (used_liquidity / self._total_liquidity) ** 2 * self._price_impact_pct
        # price_impact / self._price_impact_pct = (used_liquidity / self._total_liquidity) ** 2
        # sqrt(price_impact / self._price_impact_pct) = used_liquidity / self._total_liquidity
        # used_liquidity = self._total_liquidity * sqrt(price_impact / self._price_impact_pct)

        if price_impact == Decimal(0):
            ret = Decimal(0)
        elif self.available_liquidity == Decimal(0) or self._price_impact_pct == Decimal(0):
            raise errors.Error("Not enough liquidity")
        else:
            price_impact = min(price_impact, self._price_impact_pct)
            used_liquidity = self._total_liquidity * (price_impact / self._price_impact_pct).sqrt()

            assert used_liquidity <= self._total_liquidity
            ret = max(Decimal(0), used_liquidity - self._used_liquidity)
        return ret



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Sequence
import datetime

from basana.core import pair, event, bar
from basana.core.event_sources import csv


class RowParser(csv.RowParser):
    def __init__(
            self, pair: pair.Pair, tzinfo: datetime.tzinfo, timedelta: datetime.timedelta
    ):
        self.pair = pair
        self.tzinfo = tzinfo
        self.timedelta = timedelta

    def parse_row(self, row_dict: dict) -> Sequence[event.Event]:
        # File format:
        #
        # datetime,open,high,low,close,volume
        # 2015-01-01 00:00:00,321,321,321,321,1.73697242

        volume = Decimal(row_dict["volume"])
        # Skip bars with no volume.
        if volume == 0:
            return []

        dt = datetime.datetime.strptime(row_dict["datetime"], "%Y-%m-%d %H:%M:%S").replace(tzinfo=self.tzinfo)
        return [
            bar.BarEvent(
                dt + self.timedelta,
                bar.Bar(
                    dt, self.pair, Decimal(row_dict["open"]), Decimal(row_dict["high"]), Decimal(row_dict["low"]),
                    Decimal(row_dict["close"]), volume
                )
            )
        ]



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Set
import abc
import asyncio
import json
import logging
import time

import aiohttp

from . import event, helpers, logs


logger = logging.getLogger(__name__)


# Base class for event sources that generate events from websocket messages.
class ChannelEventSource(event.FifoQueueEventSource):
    def __init__(self, producer: event.Producer):
        super().__init__(producer=producer)

    @abc.abstractmethod
    async def push_from_message(self, message: dict):
        raise NotImplementedError()


class WebSocketClient(event.Producer, metaclass=abc.ABCMeta):
    """"Base class for channel based web socket clients."""
    def __init__(
            self, url: str, session: Optional[aiohttp.ClientSession] = None, config_overrides: dict = {},
            heartbeat: float = 30
    ):
        super().__init__()
        self._url = url
        self._session = session
        self._config_overrides = config_overrides
        self._event_sources: Dict[str, ChannelEventSource] = {}
        self._reconnect_request = asyncio.Event()
        self._subscribe_request = asyncio.Event()
        self._pending_subscriptions: Set[str] = set()
        self.backoff_secs = 1
        self._run_called = False
        self._heartbeat = heartbeat

    def set_channel_event_source(self, channel: str, event_source: ChannelEventSource):
        assert channel not in self._event_sources, "channel already registered"
        self._event_sources[channel] = event_source
        self._pending_subscriptions.add(channel)
        self._subscribe_request.set()

    def get_channel_event_source(self, channel: str) -> Optional[ChannelEventSource]:
        return self._event_sources.get(channel)

    def schedule_reconnection(self):
        self._reconnect_request.set()

    def schedule_resubscription(self, channels: List[str]):
        self._pending_subscriptions.update(channels)

    async def on_error(self, error: Any):
        logger.error(logs.StructuredMessage("Error", src=self, error=error))

    async def on_unknown_message(self, message: aiohttp.WSMessage):
        logger.warning(logs.StructuredMessage("Unknown message", src=self, type=message.type, data=message.data))

    async def main(self):
        assert not self._run_called, "run already called"

        self._run_called = True
        last_connect_ts = 0
        while True:
            # Backoff, if necessary, before connecting.
            prev_attemp_age = time.time() - last_connect_ts
            if prev_attemp_age < self.backoff_secs:
                await asyncio.sleep(self.backoff_secs - prev_attemp_age)

            try:
                logger.debug(logs.StructuredMessage("Connecting websocket", src=self, url=self._url))
                last_connect_ts = time.time()
                async with helpers.use_or_create_session(session=self._session) as session, \
                        session.ws_connect(self._url, heartbeat=self._heartbeat) as ws_cli, \
                        helpers.TaskGroup() as tg:

                    # Turn this off since we just reconnected.
                    self._reconnect_request.clear()

                    # Connect to all channels.
                    self._pending_subscriptions.update(self._event_sources.keys())
                    self._subscribe_request.set()

                    # Run tasks.
                    tg.create_task(self._msg_loop(ws_cli))
                    tg.create_task(self._subscribe_loop(ws_cli))
                    tg.create_task(self._reconnect(ws_cli))
            except Exception as e:
                await self.on_error(e)

    @abc.abstractmethod
    async def subscribe_to_channels(
            self, channels: List[str], ws_cli: aiohttp.ClientWebSocketResponse
    ):
        raise NotImplementedError()

    @abc.abstractmethod
    async def handle_message(self, message: dict) -> bool:
        raise NotImplementedError()

    async def _msg_loop(self, ws_cli: aiohttp.ClientWebSocketResponse):
        logger.debug(logs.StructuredMessage("Running message loop", src=self))

        # The iterator exits normally when the connection is closed with close code 1000 (OK) or 1001 (going away).
        # It raises a ConnectionClosedError when the connection is closed with any other code.
        async for message in ws_cli:
            handled = False
            if message.type == aiohttp.WSMsgType.TEXT:
                json_msg = json.loads(message.data)
                handled = await self.handle_message(json_msg)
            if not handled:
                await self.on_unknown_message(message)

        # If the message loop finished we need to notify the other tasks so they can finish as well.
        self._subscribe_request.set()
        self._reconnect_request.set()

    async def _reconnect(self, ws_cli: aiohttp.ClientWebSocketResponse):
        # Will exit when reconnection is requested or when its canceled.
        await self._reconnect_request.wait()
        self._reconnect_request.clear()
        # If the client is already closed then there is nothing left to do.
        if not ws_cli.closed:
            await ws_cli.close()

    async def _subscribe_loop(self, ws_cli: aiohttp.ClientWebSocketResponse):
        while not ws_cli.closed:
            await self._subscribe_request.wait()
            self._subscribe_request.clear()
            if not ws_cli.closed and self._pending_subscriptions:
                channels = list(self._pending_subscriptions)
                self._pending_subscriptions = set()
                await self.subscribe_to_channels(channels, ws_cli)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from decimal import Decimal
from typing import Any, Awaitable, Callable, List, Optional
import asyncio
import datetime
import logging

import aiohttp

from . import client, helpers
from basana.core import dt, event, logs, token_bucket, websockets as core_ws
from basana.core.pair import Pair


logger = logging.getLogger(__name__)


@dataclass
class Entry:
    #: The price.
    price: Decimal

    #: The volume.
    volume: Decimal


class OrderBook:
    """An order book."""
    def __init__(self, pair: Pair, json: dict):
        #: The trading pair.
        self.pair: Pair = pair
        #: The JSON representation.
        self.json: dict = json

    @property
    def bids(self) -> List[Entry]:
        """Returns the top bid entries."""
        return [
            Entry(price=Decimal(entry[0]), volume=Decimal(entry[1])) for entry in self.json["bids"]
        ]

    @property
    def asks(self) -> List[Entry]:
        """Returns the top ask entries."""
        return [
            Entry(price=Decimal(entry[0]), volume=Decimal(entry[1])) for entry in self.json["asks"]
        ]


class OrderBookEvent(event.Event):
    """An event for order book updates.

    :param when: The datetime when the event occurred. It must have timezone information set.
    :param order_book: The updated order book.
    """
    def __init__(self, when: datetime.datetime, order_book: OrderBook):
        super().__init__(when)
        #: The order book.
        self.order_book: OrderBook = order_book


class PollOrderBook(event.FifoQueueEventSource, event.Producer):
    def __init__(
            self, pair: Pair, interval: float, limit: Optional[int] = None,
            session: Optional[aiohttp.ClientSession] = None, tb: Optional[token_bucket.TokenBucketLimiter] = None,
            config_overrides: dict = {}
    ):
        assert interval > 0, "Invalid interval"

        super().__init__(producer=self)
        self.pair = pair
        self._interval = interval
        self._limit = limit
        self._client = client.APIClient(session=session, tb=tb, config_overrides=config_overrides)

    async def _fetch_and_push(self, order_book_symbol: str):
        order_book_json = await self._client.get_order_book(order_book_symbol, limit=self._limit)
        self.push(OrderBookEvent(
            dt.utc_now(),
            OrderBook(self.pair, order_book_json)
        ))

    async def on_error(self, error: Any):
        logger.error(logs.StructuredMessage("Error polling order book", channel=self.pair, error=error))

    async def main(self):
        order_book_symbol = helpers.pair_to_order_book_symbol(self.pair)
        while True:
            try:
                await self._fetch_and_push(order_book_symbol)
            except Exception as e:
                await self.on_error(e)
            await asyncio.sleep(self._interval)


# Generate OrderBookEvent events from websocket messages.
class WebSocketEventSource(core_ws.ChannelEventSource):
    def __init__(self, pair: Pair, producer: event.Producer):
        super().__init__(producer=producer)
        self._pair = pair

    async def push_from_message(self, message: dict):
        event = message["data"]
        self.push(OrderBookEvent(
            dt.utc_now(),  # The event doesn't include a timestamp.
            OrderBook(self._pair, event)
        ))


def get_channel(pair: Pair, depth: int) -> str:
    assert depth in [5, 10, 20], "Invalid depth"
    return "{}@depth{}".format(helpers.pair_to_order_book_symbol(pair).lower(), depth)


OrderBookEventHandler = Callable[[OrderBookEvent], Awaitable[Any]]



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional
from urllib.parse import urljoin
import abc
import asyncio
import datetime
import json
import logging
import time

import aiohttp

from . import client, config
from basana.core import dispatcher, logs, websockets as core_ws
from basana.core.config import get_config_value


logger = logging.getLogger(__name__)


class Channel(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def alias(self) -> str:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def stream(self) -> str:
        raise NotImplementedError()

    async def resolve_stream_name(self, api_client: client.APIClient):
        pass

    def keep_alive_period(self, config_overrides: dict = {}) -> Optional[datetime.timedelta]:
        return None

    async def keep_alive(self, api_client: client.APIClient):  # pragma: no cover
        pass


class PublicChannel(Channel):
    def __init__(self, name: str):
        self._name = name

    @property
    def alias(self) -> str:
        return self._name

    @property
    def stream(self) -> str:
        return self._name


class WebSocketClient(core_ws.WebSocketClient):
    def __init__(
            self, dispatcher: dispatcher.EventDispatcher, api_client: client.APIClient,
            session: Optional[aiohttp.ClientSession] = None, config_overrides: dict = {}
    ):
        url = urljoin(
            get_config_value(config.DEFAULTS, "api.websockets.base_url", overrides=config_overrides),
            "/stream"
        )
        super().__init__(
            url, session=session, config_overrides=config_overrides,
            heartbeat=get_config_value(config.DEFAULTS, "api.websockets.heartbeat", overrides=config_overrides)
        )
        self._dispatcher = dispatcher
        self._cli = api_client
        self._alias_to_channel: Dict[str, Channel] = {}
        self._stream_to_channel: Dict[str, Channel] = {}
        self._next_keep_alive: Dict[str, datetime.datetime] = {}
        self._next_msg_id = int(time.time() * 1000)

    def set_channel_event_source_ex(self, channel: Channel, event_source: core_ws.ChannelEventSource):
        assert channel.alias not in self._alias_to_channel, "channel already registered"
        super().set_channel_event_source(channel.alias, event_source)
        self._alias_to_channel[channel.alias] = channel

    def get_channel_event_source_ex(self, channel: Channel) -> Optional[core_ws.ChannelEventSource]:
        return super().get_channel_event_source(channel.alias)

    async def subscribe_to_channels(self, channel_aliases: List[str], ws_cli: aiohttp.ClientWebSocketResponse):
        logger.debug(logs.StructuredMessage("Subscribing", src=self, channels=channel_aliases))

        # Give a chance for dynamic channels to resolve the stream name.
        channels: List[Channel] = [self._alias_to_channel[alias] for alias in channel_aliases]
        await asyncio.gather(*[
            channel.resolve_stream_name(self._cli) for channel in channels
        ])
        self._stream_to_channel.update({
            channel.stream: channel for channel in channels
        })

        msg_id = self._get_next_msg_id()
        await ws_cli.send_str(json.dumps({
            "id": msg_id,
            "method": "SUBSCRIBE",
            "params": [channel.stream for channel in channels]
        }))

        # Schedule keep alives.
        for channel in channels:
            self._schedule_keep_alive(channel)

    async def handle_message(self, message: dict) -> bool:
        coro = None

        # A response to a message we sent.
        if {"result", "id"} <= set(message.keys()):
            coro = self._on_response(message)
        # A message associated to a channel.
        elif stream := message.get("stream"):
            channel = self._stream_to_channel.get(stream)
            assert channel, f"{stream} could not be mapped to a channel instance"
            # Resubscribe to the channel if the listen key expired.
            if message.get("data", {}).get("e") == "listenKeyExpired":
                logger.debug(logs.StructuredMessage(
                    "License key expired. Scheduling re-subscription", alias=channel.alias
                ))
                self.schedule_resubscription([channel.alias])
            # Get the event source for the channel alias.
            if event_source := self.get_channel_event_source(channel.alias):
                coro = event_source.push_from_message(message)

        ret = False
        if coro:
            await coro
            ret = True
        return ret

    async def _on_response(self, message: dict):
        if message["result"] is not None:
            await self.on_error(message)

    def _get_next_msg_id(self) -> int:
        ret = self._next_msg_id
        self._next_msg_id += 1
        return ret

    def _keep_alive_channel(self, channel: Channel) -> dispatcher.SchedulerJob:
        async def scheduler_job():
            if self._next_keep_alive[channel.alias] <= self._dispatcher.now():
                logger.debug(logs.StructuredMessage("Channel keep alive", alias=channel.alias))
                try:
                    await channel.keep_alive(self._cli)
                finally:
                    self._schedule_keep_alive(channel)
        return scheduler_job

    def _schedule_keep_alive(self, channel: Channel):
        period = channel.keep_alive_period(self._config_overrides)
        if period:
            schedule_dt = self._dispatcher.now() + period
            logger.debug(logs.StructuredMessage("Scheduling keep alive", when=schedule_dt, alias=channel.alias))
            self._next_keep_alive[channel.alias] = schedule_dt
            self._dispatcher.schedule(schedule_dt, self._keep_alive_channel(channel))



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional
import argparse
import datetime
import sys

import aiohttp
import asyncio

from basana.core import dt, token_bucket
from basana.external.bitstamp import client


period_to_step = {
    "min": 60,
    "hour": 3600,
    "day": 86400,
    "1m": 60,
    "3m": 180,
    "5m": 300,
    "15m": 900,
    "30m": 1800,
    "1h": 3600,
    "2h": 7200,
    "4h": 14400,
    "6h": 21600,
    "12h": 43200,
    "1d": 86400,
    "3d": 259200,
}


class OHLC:
    def __init__(self, ohlc: dict):
        self.open_timestamp = int(ohlc["timestamp"])
        self.open = ohlc["open"]
        self.high = ohlc["high"]
        self.low = ohlc["low"]
        self.close = ohlc["close"]
        self.volume = ohlc["volume"]


def parse_date(date: str):
    return datetime.datetime.combine(
        datetime.date.fromisoformat(date), datetime.time()
    ).replace(tzinfo=datetime.timezone.utc)


def to_bitstamp_currency_pair(currency_pair: str):
    parts = currency_pair.lower().split("/")
    return "".join(parts)


class CSVWriter:
    def __init__(self, output_file: Optional[str]):
        self._header_written = False
        self._output_file = open(output_file, "w") if output_file else sys.stdout

    def write_ohlc(self, ohlc: OHLC):
        if not self._header_written:
            print("datetime,open,high,low,close,volume", file=self._output_file)
            self._header_written = True

        dt_col = datetime.datetime.fromtimestamp(ohlc.open_timestamp, tz=datetime.timezone.utc)
        print(",".join([
            str(dt_col.replace(tzinfo=None)),
            ohlc.open, ohlc.high, ohlc.low, ohlc.close, ohlc.volume
        ]), file=self._output_file)


async def main(params: Optional[List[str]] = None, config_overrides: dict = {}):
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--currency-pair", help="The currency pair.", required=True)
    parser.add_argument(
        "-p", "--period", help="The period for the bars.", choices=period_to_step.keys(), required=True
    )
    parser.add_argument(
        "-s", "--start", help="The starting date YYYY-MM-DD format. Included in the range.", required=True
    )
    parser.add_argument(
        "-e", "--end", help="The ending date YYYY-MM-DD format. Included in the range.", required=True
    )
    parser.add_argument("-o", "--output", help="The output file.", required=False, default=None)
    args = parser.parse_args(args=params)

    step = period_to_step[args.period]
    start = parse_date(args.start)
    end = parse_date(args.end)
    assert start <= end, "Invalid start/end"
    # start/end are set as dates so to get past the end we need to use a step >= 1 day.
    past_the_end = end + datetime.timedelta(seconds=max(period_to_step["1d"], step))

    start_ts = dt.to_utc_timestamp(start)
    past_the_end_ts = dt.to_utc_timestamp(past_the_end)

    tb = token_bucket.TokenBucketLimiter(10, 1)
    writer = CSVWriter(args.output)
    async with aiohttp.ClientSession() as session:
        cli = client.APIClient(session=session, tb=tb, config_overrides=config_overrides)
        eof = False
        currency_pair = to_bitstamp_currency_pair(args.currency_pair)
        while not eof:
            response = await cli.get_ohlc_data(currency_pair, step, 1000, start=start_ts, exclude_current_candle=True)
            eof = True
            for ohlc in response["data"]["ohlc"]:
                eof = False
                ohlc = OHLC(ohlc)
                start_ts = max(start_ts, ohlc.open_timestamp)
                if ohlc.open_timestamp >= past_the_end_ts:
                    continue
                writer.write_ohlc(ohlc)

            if not eof:
                start_ts += step
                eof = start_ts >= past_the_end_ts


if __name__ == "__main__":  # pragma: no cover
    asyncio.run(main())



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Dict
import abc

from . import orders


class FeeStrategy(metaclass=abc.ABCMeta):
    """Base class for strategies that model fee schemes.

    .. note::

        * This is a base class and should not be used directly.
    """

    @abc.abstractmethod
    def calculate_fees(
            self, order: orders.Order, balance_updates: Dict[str, Decimal]
    ) -> Dict[str, Decimal]:
        raise NotImplementedError()


class NoFee(FeeStrategy):
    """This strategy applies no fees to the trades."""

    def calculate_fees(self, order: orders.Order, balance_updates: Dict[str, Decimal]) -> Dict[str, Decimal]:
        return {}


class Percentage(FeeStrategy):
    """This strategy applies a fixed percentage per trade, in quote currency.

    :param percentage: The percentage to apply.
    :param min_fee: Minimum fee amount, in quote currency.
    """

    def __init__(self, percentage: Decimal, min_fee: Decimal = Decimal(0)):
        assert percentage >= 0 and percentage < 100, f"Invalid percentage {percentage}"
        assert min_fee >= 0, f"Minimum fee cannot be negative {min_fee}"
        self._percentage = percentage
        self._min_fee = min_fee

    def calculate_fees(self, order: orders.Order, balance_updates: Dict[str, Decimal]) -> Dict[str, Decimal]:
        ret = {}

        # Fees are always charged in quote amount.
        symbol = order.pair.quote_symbol

        # Rounding may have taken place in previous fills, so fees may have been overcharged. For that reason we
        # calculate the total fees to charge, and subtract what we have charged so far.
        charged_fee_amount = order.fees.get(symbol, Decimal(0))
        assert charged_fee_amount <= Decimal(0), "Fees should always be negative"
        total_quote_amount = order.balance_updates.get(symbol, Decimal(0)) + balance_updates.get(symbol, Decimal(0))
        total_fee_amount = -max(abs(total_quote_amount) * self._percentage / Decimal(100), self._min_fee)
        pending_fee = total_fee_amount - charged_fee_amount
        if pending_fee < Decimal(0):
            ret[symbol] = pending_fee

        return ret



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Dict
import abc
import dataclasses
import datetime

from basana.backtesting import account_balances, config, errors, prices
from basana.backtesting.value_map import ValueMap, ValueMapDict
from basana.core import dispatcher


@dataclasses.dataclass
class LoanInfo:
    #: The loan id.
    id: str
    #: True if the loan is open, False otherwise.
    is_open: bool
    #: The symbol being borrowed.
    borrowed_symbol: str
    #: The amount being borrowed.
    borrowed_amount: Decimal
    #: The outstanding interest. Only valid for open loans.
    outstanding_interest: Dict[str, Decimal]
    #: The paid interest. Only valid for closed loans.
    paid_interest: Dict[str, Decimal]


class Loan(metaclass=abc.ABCMeta):
    def __init__(
            self, id: str, borrowed_symbol: str,  borrowed_amount: Decimal, created_at: datetime.datetime
    ):
        assert borrowed_amount > Decimal(0), f"Invalid amount {borrowed_amount}"

        self._id = id
        self._borrowed_symbol = borrowed_symbol
        self._borrowed_amount = borrowed_amount
        self._is_open = True
        self._created_at = created_at
        self._paid_interest = ValueMap()

    @property
    def id(self) -> str:
        return self._id

    @property
    def is_open(self) -> bool:
        return self._is_open

    @property
    def borrowed_symbol(self) -> str:
        return self._borrowed_symbol

    @property
    def borrowed_amount(self) -> Decimal:
        return self._borrowed_amount

    @property
    def created_at(self) -> datetime.datetime:
        return self._created_at

    @property
    def paid_interest(self) -> ValueMapDict:
        return self._paid_interest

    def close(self):
        assert self._is_open
        self._is_open = False

    def add_paid_interest(self, interest: ValueMapDict):
        self._paid_interest += interest

    @abc.abstractmethod
    def calculate_interest(self, at: datetime.datetime, prices: prices.Prices) -> ValueMapDict:
        raise NotImplementedError()

    @abc.abstractmethod
    def calculate_collateral(self, prices: prices.Prices) -> ValueMapDict:
        raise NotImplementedError()


@dataclasses.dataclass
class ExchangeContext:
    dispatcher: dispatcher.BacktestingDispatcher
    account_balances: account_balances.AccountBalances
    prices: prices.Prices
    config: config.Config


class LendingStrategy(metaclass=abc.ABCMeta):
    """
    Base class for lending strategies.
    """

    def set_exchange_context(self, loan_mgr, exchange_context: ExchangeContext):
        """
        This method will be called during exchange initialization to give lending strategies a chance to later
        use those services.
        """
        pass

    @abc.abstractmethod
    def create_loan(self, symbol: str, amount: Decimal, created_at: datetime.datetime) -> Loan:
        raise NotImplementedError()


class NoLoans(LendingStrategy):
    """
    Lending not supported.
    """

    def create_loan(self, symbol: str, amount: Decimal, created_at: datetime.datetime) -> Loan:
        raise errors.Error("Lending is not supported")



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Optional, Dict
import datetime

from . import client, config, margin, user_data, websockets, websocket_mgr
from .client import margin as margin_client
from basana.core.config import get_config_value


# Forward declarations
OrderEvent = user_data.OrderEvent
OrderEventHandler = user_data.OrderEventHandler
OrderUpdate = user_data.OrderUpdate
UserDataEvent = user_data.Event
UserDataEventHandler = user_data.UserDataEventHandler


class CrossMarginUserDataChannel(websockets.Channel):
    def __init__(self):
        self._listen_key = None

    @property
    def alias(self) -> str:
        return "cross_margin_user_data"

    @property
    def stream(self) -> str:
        assert self._listen_key, "resolve_stream_name not called"
        return self._listen_key

    async def resolve_stream_name(self, api_client: client.APIClient):
        self._listen_key = (await api_client.cross_margin_account.create_listen_key())["listenKey"]

    def keep_alive_period(self, config_overrides: dict = {}) -> Optional[datetime.timedelta]:
        return datetime.timedelta(
            seconds=get_config_value(
                config.DEFAULTS, "api.websockets.cross_margin.user_data_stream.heartbeat", overrides=config_overrides
            )
        )

    async def keep_alive(self, api_client: client.APIClient):
        assert self._listen_key, "resolve_stream_name not called"
        await api_client.cross_margin_account.keep_alive_listen_key(self._listen_key)


class Account(margin.Account):
    """Cross margin account."""
    def __init__(self, cli: margin_client.CrossMarginAccount, ws_mgr: websocket_mgr.WebsocketManager):
        self._cli = cli
        self._ws_mgr = ws_mgr

    @property
    def client(self) -> margin_client.CrossMarginAccount:
        return self._cli

    async def get_balances(self) -> Dict[str, margin.Balance]:
        """Returns all balances."""
        account_info = await self.client.get_account_information()
        return {balance["asset"].upper(): margin.Balance(balance) for balance in account_info["userAssets"]}

    async def transfer_from_spot_account(self, asset: str, amount: Decimal) -> dict:
        """Transfer balances from the spot account to the cross margin account.

        If the transfer can't be completed a :class:`basana.external.binance.exchange.Error` will be raised.

        :param asset: The asset to transfer.
        :param amount: The amount to transfer.
        """
        return await self.client.transfer_from_spot_account(asset, amount)

    async def transfer_to_spot_account(self, asset: str, amount: Decimal) -> dict:
        """Transfer balances from the cross margin account to the spot account.

        If the transfer can't be completed a :class:`basana.external.binance.exchange.Error` will be raised.

        :param asset: The asset to transfer.
        :param amount: The amount to transfer.
        """
        return await self.client.transfer_to_spot_account(asset, amount)

    def subscribe_to_user_data_events(self, event_handler: UserDataEventHandler):
        """
        Registers an async callable that will be called for every new user data event.

        Works as defined in https://developers.binance.com/docs/margin_trading/trade-data-stream.

        :param event_handler: The event handler.
        """

        self._ws_mgr.subscribe_to_user_data_events(
            CrossMarginUserDataChannel(),
            lambda ws_cli: user_data.WebSocketEventSource(ws_cli),
            event_handler
        )

    def subscribe_to_order_events(self, event_handler: OrderEventHandler):
        """
        Registers an async callable that will be called for every new order update.

        Works as defined in https://developers.binance.com/docs/margin_trading/trade-data-stream/Event-Order-Update.

        :param event_handler: The event handler.
        """

        self._ws_mgr.subscribe_to_order_events(
            CrossMarginUserDataChannel(),
            lambda ws_cli: user_data.WebSocketEventSource(ws_cli),
            event_handler
        )



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional
import abc
import datetime

from . import dt


class Producer:
    """Base class for producers.

    A producer is the active part of an :class:`basana.EventSource` or a group of them.
    Take a look at :meth:`EventDispatcher.run` for details on how producers are used.

    .. note::

        This is a base class and should not be used directly.
    """

    async def initialize(self):
        """Override to perform initialization."""
        pass

    async def main(self):
        """Override to run the loop that produces events."""
        pass

    async def finalize(self):
        """Override to perform cleanup."""
        pass


class Event:
    """Base class for events.

    An event is something that occurs at a specific point in time. There are many different types of events:

    * An update to an order book.
    * A new trade.
    * An order update.
    * A new bar (candlestick/ohlc).
    * Others

    :param when: The datetime when the event occurred. It must have timezone information set.

    .. note::

        This is a base class and should not be used directly.
    """

    def __init__(self, when: datetime.datetime):
        assert not dt.is_naive(when), f"{when} should have timezone information set"

        #: The datetime when the event occurred.
        self.when: datetime.datetime = when


class EventSource(metaclass=abc.ABCMeta):
    """Base class for event sources.

    This class declares the interface that is required by the :class:`basana.EventDispatcher` to gather events for
    processing.

    :param producer: An optional producer associated with this event source.
    """

    def __init__(self, producer: Optional[Producer] = None):
        self.producer = producer

    @abc.abstractmethod
    def pop(self) -> Optional[Event]:
        """Override to return the next event, or None if there are no events available.

        This method is used by the :class:`basana.EventDispatcher` during the event dispatch loop so **it should return
        as soon as possible**.
        """
        raise NotImplementedError()


class FifoQueueEventSource(EventSource):
    """A FIFO queue event source.

    :param producer: An optional producer associated with this event source.
    :param events: An optional list of initial events.
    """
    def __init__(self, producer: Optional[Producer] = None, events: List[Event] = []):
        super().__init__(producer)
        self._queue: List[Event] = []
        self._queue.extend(events)

    def push(self, event: Event):
        """Adds an event to the end of the queue."""
        self._queue.append(event)

    def pop(self) -> Optional[Event]:
        """Removes and returns the next event in the queue."""
        ret = None
        if self._queue:
            ret = self._queue.pop(0)
        return ret



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import cast, Any, Awaitable, Callable, Dict, Iterable, List, Optional, Tuple, Union
import datetime

from basana.core import dispatcher, enums, errors, event, helpers, pair


class BaseTradingSignal(event.Event):
    def __init__(self, when: datetime.datetime):
        super().__init__(when)
        self._positions: Dict[pair.Pair, enums.Position] = {}

    def add_pair(self, pair: pair.Pair, position: enums.Position):
        """
        Add a pair to the signal.

        :param pair: The pair to trade.
        :param position: The position to switch to.
        """
        self._positions[pair] = position

    def get_pairs(self) -> Iterable[Tuple[pair.Pair, enums.Position]]:
        return self._positions.items()

    def get_position(self, pair: pair.Pair) -> enums.Position:
        return self._positions[pair]


class TradingSignal(BaseTradingSignal):
    """
    A trading signal is an event that instructs to take a long, short, or neutral position on a given trading pair.

    :param when: The datetime when the trading signal occurred. It must have timezone information set.
    :param op_or_pos: A enums.Position or an enums.OrderOperation (for backwards compatibility purposes).
    :param pair: The pair to trade.
    """

    def __init__(
            self, when: datetime.datetime, op_or_pos: Union[enums.OrderOperation, enums.Position], pair: pair.Pair
    ):
        super().__init__(when)

        if isinstance(op_or_pos, enums.OrderOperation):
            helpers.deprecation_warning(
                "Support for bs.OrderOperation in trading signals will be removed soon."
                " Switch to bs.Position"
            )
            op_or_pos = {
                enums.OrderOperation.BUY: enums.Position.LONG,
                enums.OrderOperation.SELL: enums.Position.SHORT,
            }[op_or_pos]
        self.add_pair(pair, op_or_pos)

    @property
    def pair(self) -> pair.Pair:
        """
        The pair to trade.
        """
        pair, _ = next(iter(self.get_pairs()))
        return pair

    @property
    def position(self) -> enums.Position:
        """
        The position to switch to.
        """
        return self.get_position(self.pair)

    @property
    def operation(self) -> enums.OrderOperation:
        """
        The operation.

        .. note::

          * This property is deprecated and position should be used instead.
        """
        position = self.position
        op = {
            enums.Position.LONG: enums.OrderOperation.BUY,
            enums.Position.SHORT: enums.OrderOperation.SELL,
        }.get(position)
        if op is None:
            raise errors.Error("{} can't be mapped to an operation".format(position))
        return op


class TradingSignalSource(event.FifoQueueEventSource):
    """
    Base class for event sources that generate :class:`basana.BaseTradingSignal` events.

    :param dispatcher: The event dispatcher.
    :param producer: An optional producer associated with this event source.
    :param events: An optional list of initial events.
    """

    def __init__(
            self, dispatcher: dispatcher.EventDispatcher, producer: Optional[event.Producer] = None,
            events: List[event.Event] = []
    ):
        super().__init__(producer=producer, events=events)
        self._dispatcher = dispatcher

    def subscribe_to_trading_signals(self, event_handler: Callable[[BaseTradingSignal], Awaitable[Any]]):
        """
        Registers an async callable that will be called when a new trading signal is available.

        :param event_handler: An async callable that receives an trading signal.
        """

        self._dispatcher.subscribe(self, cast(dispatcher.EventHandler, event_handler))



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Dict, Optional, Sequence, Tuple
from urllib.parse import urljoin
import asyncio
import copy
import time

import aiohttp

from basana.core import token_bucket, helpers as core_helpers
from basana.core.config import get_config_value
from basana.external.binance import config, helpers


class Error(Exception):
    """
    An error returned by the exchange.

    :param msg: The error message, if available.
    :param code: The error code, if available.
    :param resp: The response.
    :param json_response: The response body, if it was a JSON.
    """
    def __init__(self, msg: str, code: Optional[int], resp: aiohttp.ClientResponse, json_response: Optional[Any]):
        super().__init__(msg)
        #: The error message.
        self.msg = msg
        #: The error code, if available.
        self.code = code
        #: The HTTP status code.
        self.http_status = resp.status
        #: The HTTP reason.
        self.http_reason = resp.reason
        #: The response body, if it was a JSON.
        self.json_response = json_response


def raise_for_error(resp: aiohttp.ClientResponse, json_response):
    msg = None
    code = None
    if isinstance(json_response, dict):
        msg = json_response.get("msg")
        code = json_response.get("code")
    if msg is None and not resp.ok:
        msg = "{} {}".format(resp.status, resp.reason)
    if msg is not None:
        raise Error(msg, code, resp, json_response)


class BaseClient:
    def __init__(
            self, api_key: Optional[str] = None, api_secret: Optional[str] = None,
            session: Optional[aiohttp.ClientSession] = None, tb: Optional[token_bucket.TokenBucketLimiter] = None,
            config_overrides: dict = {}
    ):
        assert not ((api_key is None) ^ (api_secret is None)), \
            "Both api_key and api_secret should be set, or none of them"

        self._api_key = api_key
        self._api_secret = api_secret
        self._session = session
        self._tb = tb
        self._config_overrides = config_overrides

    async def make_request(
            self, method: str, path: str, send_key: bool = False, send_sig: bool = False,
            qs_params: Dict[str, Any] = {}, data: Dict[str, Any] = {}
    ) -> Any:
        if self._tb and (sleep_time := self._tb.consume()):
            await asyncio.sleep(sleep_time)

        async with core_helpers.use_or_create_session(session=self._session) as session:
            headers = {}
            session_method = {
                "DELETE": session.delete,
                "GET": session.get,
                "POST": session.post,
                "PUT": session.put,
            }.get(method)
            assert session_method is not None

            base_url = get_config_value(config.DEFAULTS, "api.http.base_url", overrides=self._config_overrides)
            timeout = get_config_value(config.DEFAULTS, "api.http.timeout", overrides=self._config_overrides)
            url = urljoin(base_url, path)

            if send_key or send_sig:
                assert self._api_key, "api_key not set"

                headers["X-MBX-APIKEY"] = self._api_key

            if send_sig:
                assert self._api_secret, "api_secret not set"

                qs_params = copy.copy(qs_params)
                # Signature and timestamp should go in the query string, and the timestamp should be included in the
                # signature.
                qs_params["timestamp"] = int(round(time.time() * 1000))
                qs_params["signature"] = helpers.get_signature(self._api_secret, qs_params=qs_params, data=data)

            form_data = None if not data else aiohttp.FormData(data)
            async with session_method(url, headers=headers, params=qs_params, data=form_data, timeout=timeout) as resp:
                # print(await resp.text())
                json_response = None
                if (ct := resp.headers.get("Content-Type")) and ct.lower().find("application/json") == 0:
                    json_response = await resp.json()
                raise_for_error(resp, json_response)
                return json_response


def set_optional_params(params: Dict[str, Any], tuples: Sequence[Tuple[str, Any]]):
    for k, v in tuples:
        if v is None:
            continue
        if isinstance(v, Decimal):
            v = str(v)
        params[k] = v
